import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Mapping, Callable
import io
import PIL.Image
from autogen_core import Image as AGImage
import asyncio
from autogen_core import (
    CancellationToken,
    DefaultTopicId,
    MessageContext,
    event,
    rpc,
    AgentId,
)
from autogen_core.models import (
    ChatCompletionClient,
    LLMMessage,
    SystemMessage,
    UserMessage,
)
from autogen_core.model_context import TokenLimitedChatCompletionContext
from autogen_agentchat.base import Response, TerminationCondition
from autogen_agentchat.messages import (
    BaseChatMessage,
    StopMessage,
    TextMessage,
    MultiModalMessage,
    BaseAgentEvent,
    MessageFactory,
)
from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatMessage,
    GroupChatRequestPublish,
    GroupChatStart,
    GroupChatTermination,
)
from autogen_agentchat.teams._group_chat._base_group_chat_manager import (
    BaseGroupChatManager,
)
from autogen_agentchat.state import BaseGroupChatManagerState
from .....ui_backend.learning.memory_provider import MemoryControllerProvider

from .....ui_backend.types import HumanInputFormat, Plan
from .....ui_backend.utils import dict_to_str, thread_to_context
from ...tools.bing_search import get_bing_search_results
from ...teams.orchestrator.orchestrator_config import OrchestratorConfig
from ._prompts import (
    ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING,
    ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING_AUTONOMOUS,
    ORCHESTRATOR_SYSTEM_MESSAGE_EXECUTION,
    ORCHESTRATOR_FINAL_ANSWER_PROMPT,
    ORCHESTRATOR_PROGRESS_LEDGER_PROMPT,
    ORCHESTRATOR_TASK_LEDGER_FULL_FORMAT,
    ORCHESTRATOR_PLAN_PROMPT_JSON,
    ORCHESTRATOR_PLAN_REPLAN_JSON,
    INSTRUCTION_AGENT_FORMAT,
    validate_ledger_json,
    validate_plan_json,
)
from ._utils import is_accepted_str, extract_json_from_string
from loguru import logger as trace_logger


class OrchestratorState(BaseGroupChatManagerState):
    """
    The OrchestratorState class is responsible for maintaining the state of the group chat conversation.
    """

    task: str = ""
    plan_str: str = ""
    plan: Plan | None = None
    n_rounds: int = 0
    current_step_idx: int = 0
    information_collected: str = ""
    in_planning_mode: bool = True
    is_paused: bool = False
    group_topic_type: str = ""
    message_history: List[BaseChatMessage | BaseAgentEvent] = []
    participant_topic_types: List[str] = []
    n_replans: int = 0

    def reset(self) -> None:
        self.task = ""
        self.plan_str = ""
        self.plan = None
        self.n_rounds = 0
        self.current_step_idx = 0
        self.information_collected = ""
        self.in_planning_mode = True
        self.message_history = []
        self.is_paused = False
        self.n_replans = 0

    def reset_for_followup(self) -> None:
        self.task = ""
        self.plan_str = ""
        self.plan = None
        self.n_rounds = 0
        self.current_step_idx = 0
        self.in_planning_mode = True
        self.is_paused = False
        self.n_replans = 0


class Orchestrator(BaseGroupChatManager):
    """
    The Orchestrator class is responsible for managing a group chat by orchestrating the conversation
    between multiple participants. It extends the SequentialRoutedAgent class and provides functionality
    to handle the start, reset, and progression of the group chat.

    The orchestrator maintains the state of the conversation, including the task, plan, and progress. It
    interacts with a model client to generate and validate plans, and it publishes messages to the group
    chat based on the current state and responses from participants.

    """

    def __init__(
        self,
        name: str,
        group_topic_type: str,
        output_topic_type: str,
        message_factory: MessageFactory,
        participant_topic_types: List[str],
        participant_descriptions: List[str],
        participant_names: List[str],
        output_message_queue: asyncio.Queue[
            BaseAgentEvent | BaseChatMessage | GroupChatTermination
        ],
        model_client: ChatCompletionClient,
        config: OrchestratorConfig,
        termination_condition: TerminationCondition | None = None,
        max_turns: int | None = None,
        memory_provider: MemoryControllerProvider | None = None,
    ):
        super().__init__(
            name,
            group_topic_type,
            output_topic_type,
            participant_topic_types,
            participant_names,
            participant_descriptions,
            output_message_queue,
            termination_condition,
            max_turns,
            message_factory=message_factory,
        )
        self._model_client: ChatCompletionClient = model_client
        self._model_context = TokenLimitedChatCompletionContext(
            model_client, token_limit=config.model_context_token_limit
        )
        self._config: OrchestratorConfig = config
        self._user_agent_topic = "user_proxy"
        self._web_agent_topic = "web_surfer"
        if self._user_agent_topic not in self._participant_names:
            if not (
                self._config.autonomous_execution
                and not self._config.allow_follow_up_input
            ):
                raise ValueError(
                    f"User agent topic {self._user_agent_topic} not in participant names {self._participant_names}"
                )

        self._memory_controller = None
        self._memory_provider = memory_provider
        if (
            self._config.memory_controller_key
            and self._model_client
            and self._memory_provider is not None
        ):
            try:
                provider = self._memory_provider
                self._memory_controller = provider.get_memory_controller(
                    memory_controller_key=self._config.memory_controller_key,
                    client=self._model_client,
                )
                trace_logger.info("Memory controller initialized successfully.")
            except Exception as e:
                trace_logger.warning(f"Failed to initialize memory controller: {e}")

        # Setup internal variables
        self._setup_internals()

    def _setup_internals(self) -> None:
        """
        Setup internal variables used in orchestrator
        """
        self._state: OrchestratorState = OrchestratorState()

        # Create filtered lists for execution that may exclude the user agent
        self._agent_execution_names = self._participant_names.copy()
        self._agent_execution_descriptions = self._participant_descriptions.copy()

        if self._config.autonomous_execution:
            # Filter out the user agent from execution lists
            user_indices = [
                i
                for i, name in enumerate(self._agent_execution_names)
                if name == self._user_agent_topic
            ]
            if user_indices:
                user_index = user_indices[0]
                self._agent_execution_names.pop(user_index)
                self._agent_execution_descriptions.pop(user_index)
        # add a a new participant for the orchestrator to do nothing
        self._agent_execution_names.append("no_action_agent")
        self._agent_execution_descriptions.append(
            "If for this step no action is needed, you can use this agent to perform no action"
        )

        self._team_description: str = "\n".join(
            [
                f"{topic_type}: {description}".strip()
                for topic_type, description in zip(
                    self._agent_execution_names,
                    self._agent_execution_descriptions,
                    strict=True,
                )
            ]
        )
        self._last_browser_metadata_hash = ""

    def _get_system_message_planning(
        self,
    ) -> str:
        date_today = datetime.now().strftime("%Y-%m-%d")
        if self._config.autonomous_execution:
            return ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING_AUTONOMOUS.format(
                date_today=date_today,
                team=self._team_description,
            )
        else:
            return ORCHESTRATOR_SYSTEM_MESSAGE_PLANNING.format(
                date_today=date_today,
                team=self._team_description,
            )

    def _get_task_ledger_plan_prompt(self, team: str) -> str:
        additional_instructions = ""
        if self._config.allowed_websites is not None:
            additional_instructions = (
                "Only use the following websites if possible: "
                + ", ".join(self._config.allowed_websites)
            )

        return ORCHESTRATOR_PLAN_PROMPT_JSON.format(
            team=team, additional_instructions=additional_instructions
        )

    def _get_task_ledger_replan_plan_prompt(
        self, task: str, team: str, plan: str
    ) -> str:
        additional_instructions = ""
        if self._config.allowed_websites is not None:
            additional_instructions = (
                "Only use the following websites if possible: "
                + ", ".join(self._config.allowed_websites)
            )
        return ORCHESTRATOR_PLAN_REPLAN_JSON.format(
            task=task,
            team=team,
            plan=plan,
            additional_instructions=additional_instructions,
        )

    def _get_task_ledger_full_prompt(self, task: str, team: str, plan: str) -> str:
        return ORCHESTRATOR_TASK_LEDGER_FULL_FORMAT.format(
            task=task, team=team, plan=plan
        )

    def _get_progress_ledger_prompt(
        self, task: str, plan: str, step_index: int, team: str, names: List[str]
    ) -> str:
        assert self._state.plan is not None
        additional_instructions = ""
        if self._config.autonomous_execution:
            additional_instructions = "VERY IMPORTANT: The next agent name cannot be the user or user_proxy, use any other agent."
        return ORCHESTRATOR_PROGRESS_LEDGER_PROMPT.format(
            task=task,
            plan=plan,
            step_index=step_index,
            step_title=self._state.plan[step_index].title,
            step_details=self._state.plan[step_index].details,
            agent_name=self._state.plan[step_index].agent_name,
            team=team,
            names=", ".join(names),
            additional_instructions=additional_instructions,
        )

    def _get_final_answer_prompt(self, task: str) -> str:
        if self._config.final_answer_prompt is not None:
            return self._config.final_answer_prompt.format(task=task)
        else:
            return ORCHESTRATOR_FINAL_ANSWER_PROMPT.format(task=task)

    def get_agent_instruction(self, instruction: str, agent_name: str) -> str:
        assert self._state.plan is not None
        return INSTRUCTION_AGENT_FORMAT.format(
            step_index=self._state.current_step_idx + 1,
            step_title=self._state.plan[self._state.current_step_idx].title,
            step_details=self._state.plan[self._state.current_step_idx].details,
            agent_name=agent_name,
            instruction=instruction,
        )

    def _validate_ledger_json(self, json_response: Dict[str, Any]) -> bool:
        return validate_ledger_json(json_response, self._agent_execution_names)

    def _validate_plan_json(self, json_response: Dict[str, Any]) -> bool:
        return validate_plan_json(json_response)

    async def validate_group_state(
        self, messages: List[BaseChatMessage] | None
    ) -> None:
        pass

    async def select_speaker(
        self, thread: List[BaseAgentEvent | BaseChatMessage]
    ) -> str:
        """Not used in this class."""
        return ""

    async def reset(self) -> None:
        """Reset the group chat manager."""
        if self._termination_condition is not None:
            await self._termination_condition.reset()
        self._state.reset()

    async def _log_message(self, log_message: str) -> None:
        trace_logger.debug(log_message)

    async def _log_message_agentchat(
        self,
        content: str,
        internal: bool = False,
        metadata: Optional[Dict[str, str]] = None,
    ) -> None:
        internal_str = "yes" if internal else "no"
        message = TextMessage(
            content=content,
            source=self._name,
            metadata=metadata or {"internal": internal_str},
        )
        await self.publish_message(
            GroupChatMessage(message=message),
            topic_id=DefaultTopicId(type=self._output_topic_type),
        )
        await self._output_message_queue.put(message)

    async def _publish_group_chat_message(
        self,
        content: str,
        cancellation_token: CancellationToken,
        internal: bool = False,
        metadata: Optional[Dict[str, str]] = None,
    ) -> None:
        """Helper function to publish a group chat message."""
        internal_str = "yes" if internal else "no"
        message = TextMessage(
            content=content,
            source=self._name,
            metadata=metadata or {"internal": internal_str},
        )
        await self.publish_message(
            GroupChatMessage(message=message),
            topic_id=DefaultTopicId(type=self._output_topic_type),
        )
        await self._output_message_queue.put(message)
        await self.publish_message(
            GroupChatAgentResponse(agent_response=Response(chat_message=message)),
            topic_id=DefaultTopicId(type=self._group_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _request_next_speaker(
        self, next_speaker: str, cancellation_token: CancellationToken
    ) -> None:
        """Helper function to request the next speaker."""
        if next_speaker == "no_action_agent":
            await self._orchestrate_step(cancellation_token)
            return

        next_speaker_topic_type = self._participant_name_to_topic_type[next_speaker]
        await self.publish_message(
            GroupChatRequestPublish(),
            topic_id=DefaultTopicId(type=next_speaker_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _get_json_response(
        self,
        messages: List[LLMMessage],
        validate_json: Callable[[Dict[str, Any]], bool],
        cancellation_token: CancellationToken,
    ) -> Dict[str, Any] | None:
        """Get a JSON response from the model client.
        Args:
            messages (List[LLMMessage]): The messages to send to the model client.
            validate_json (callable): A function to validate the JSON response. The function should return True if the JSON response is valid, otherwise False.
            cancellation_token (CancellationToken): A token to cancel the request if needed.
        """
        retries = 0
        exception_message = ""
        while retries < self._config.max_json_retries:
            # Re-initialize model context to meet token limit quota
            await self._model_context.clear()
            for msg in messages:
                await self._model_context.add_message(msg)
            if exception_message != "":
                await self._model_context.add_message(
                    UserMessage(content=exception_message, source=self._name)
                )
            token_limited_messages = await self._model_context.get_messages()

            response = await self._model_client.create(
                token_limited_messages,
                json_output=True
                if self._model_client.model_info["json_output"]
                else False,
                cancellation_token=cancellation_token,
            )
            assert isinstance(response.content, str)
            try:
                json_response = json.loads(response.content)
                # Use the validate_json function to check the response
                if validate_json(json_response):
                    return json_response
                else:
                    exception_message = "Validation failed for JSON response, retrying. You must return a valid JSON object parsed from the response."
                    await self._log_message(
                        f"Validation failed for JSON response, retrying ({retries}/{self._config.max_json_retries})"
                    )
            except json.JSONDecodeError as e:
                json_response = extract_json_from_string(response.content)
                if json_response is not None:
                    if validate_json(json_response):
                        return json_response
                    else:
                        exception_message = "Validation failed for JSON response, retrying. You must return a valid JSON object parsed from the response."
                else:
                    exception_message = f"Failed to parse JSON response, retrying. You must return a valid JSON object parsed from the response. Error: {e}"
                await self._log_message(
                    f"Failed to parse JSON response, retrying ({retries}/{self._config.max_json_retries})"
                )
            retries += 1
        raise ValueError("Failed to get a valid JSON response after multiple retries")

    @rpc
    async def handle_start(self, message: GroupChatStart, ctx: MessageContext) -> None:  # type: ignore
        """Handle the start of a group chat by selecting a speaker to start the conversation."""
        # Check if the conversation has already terminated.
        if (
            self._termination_condition is not None
            and self._termination_condition.terminated
        ):
            early_stop_message = StopMessage(
                content="The group chat has already terminated.", source=self._name
            )
            await self._signal_termination(early_stop_message)

            # Stop the group chat.
            return
        assert message is not None and message.messages is not None

        # send message to all agents with initial user message
        await self.publish_message(
            GroupChatStart(messages=message.messages),
            topic_id=DefaultTopicId(type=self._group_topic_type),
            cancellation_token=ctx.cancellation_token,
        )

        # handle agent response
        for m in message.messages:
            self._state.message_history.append(m)
        await self._orchestrate_step(ctx.cancellation_token)

    async def pause(self) -> None:
        """Pause the group chat manager."""
        self._state.is_paused = True

    async def resume(self) -> None:
        """Resume the group chat manager."""
        self._state.is_paused = False

    @event
    async def handle_agent_response(
        self, message: GroupChatAgentResponse, ctx: MessageContext
    ) -> None:  # type: ignore
        delta: List[BaseChatMessage] = []
        if message.agent_response.inner_messages is not None:
            for inner_message in message.agent_response.inner_messages:
                delta.append(inner_message)  # type: ignore
        self._state.message_history.append(message.agent_response.chat_message)
        delta.append(message.agent_response.chat_message)

        if self._termination_condition is not None:
            stop_message = await self._termination_condition(delta)
            if stop_message is not None:
                await self._prepare_final_answer(
                    reason="Termination Condition Met.",
                    cancellation_token=ctx.cancellation_token,
                    force_stop=True,
                )
                await self._signal_termination(stop_message)
                # Stop the group chat and reset the termination conditions and turn count.
                await self._termination_condition.reset()
                return
        await self._orchestrate_step(ctx.cancellation_token)

    async def _orchestrate_step(self, cancellation_token: CancellationToken) -> None:
        """Orchestrate the next step of the conversation."""
        if self._state.is_paused:
            # let user speak next if paused
            await self._request_next_speaker(self._user_agent_topic, cancellation_token)
            return

        if self._state.in_planning_mode:
            await self._orchestrate_step_planning(cancellation_token)
        else:
            await self._orchestrate_step_execution(cancellation_token)

    async def do_bing_search(self, query: str) -> str | None:
        try:
            # log the bing search request
            await self._log_message_agentchat(
                "Searching online for information...",
                metadata={"internal": "no", "type": "progress_message"},
            )
            bing_search_results = await get_bing_search_results(
                query,
                max_pages=3,
                max_tokens_per_page=5000,
                timeout_seconds=35,
            )
            if bing_search_results.combined_content != "":
                bing_results_progress = f"Reading through {len(bing_search_results.page_contents)} web pages..."
                await self._log_message_agentchat(
                    bing_results_progress,
                    metadata={"internal": "no", "type": "progress_message"},
                )
                return bing_search_results.combined_content
            return None
        except Exception as e:
            trace_logger.exception(f"Error in doing bing search: {e}")
            return None

    async def _get_websurfer_page_info(self) -> None:
        """Get the page information from the web surfer agent."""
        try:
            if self._web_agent_topic in self._participant_names:
                web_surfer_container = (
                    await self._runtime.try_get_underlying_agent_instance(
                        AgentId(
                            type=self._participant_name_to_topic_type[
                                self._web_agent_topic
                            ],
                            key=self.id.key,
                        )
                    )
                )
                if (
                    web_surfer_container._agent is not None  # type: ignore
                ):
                    web_surfer = web_surfer_container._agent  # type: ignore
                    page_title: str | None = None
                    page_url: str | None = None
                    (page_title, page_url) = await web_surfer.get_page_title_url()  # type: ignore
                    assert page_title is not None
                    assert page_url is not None

                    num_tabs, tabs_information_str = await web_surfer.get_tabs_info()  # type: ignore
                    tabs_information_str = f"There are {num_tabs} tabs open. The tabs are as follows:\n{tabs_information_str}"

                    message = MultiModalMessage(
                        content=[tabs_information_str],
                        source="web_surfer",
                    )
                    if "about:blank" not in page_url:
                        page_description: str | None = None
                        screenshot: bytes | None = None
                        metadata_hash: str | None = None
                        (
                            page_description,  # type: ignore
                            screenshot,  # type: ignore
                            metadata_hash,  # type: ignore
                        ) = await web_surfer.describe_current_page()  # type: ignore
                        assert isinstance(screenshot, bytes)
                        assert isinstance(page_description, str)
                        assert isinstance(metadata_hash, str)
                        if metadata_hash != self._last_browser_metadata_hash:
                            page_description = (
                                "A description of the current page: " + page_description
                            )
                            self._last_browser_metadata_hash: str = metadata_hash

                            message.content.append(page_description)
                            message.content.append(
                                AGImage.from_pil(PIL.Image.open(io.BytesIO(screenshot)))
                            )
                    self._state.message_history.append(message)
        except Exception as e:
            trace_logger.exception(f"Error in getting web surfer screenshot: {e}")
            pass

    async def _handle_relevant_plan_from_memory(
        self,
        context: Optional[List[LLMMessage]] = None,
    ) -> Optional[Any]:
        """
        Handles retrieval of relevant plans from memory for 'reuse' or 'hint' modes.
        Returns:
            For 'reuse', returns the most relevant plan (or None).
            For 'hint', appends a relevant plan as a UserMessage to the context if found.
        """
        if not self._memory_controller:
            return None
        try:
            mode = self._config.retrieve_relevant_plans
            task = self._state.task
            source = self._name
            trace_logger.info(
                f"retrieving relevant plan from memory for mode: {mode} ..."
            )
            memos = await self._memory_controller.retrieve_relevant_memos(task=task)
            trace_logger.info(f"{len(memos)} relevant plan(s) retrieved from memory")
            if len(memos) > 0:
                most_relevant_plan = memos[0].insight
                if mode == "reuse":
                    return most_relevant_plan
                elif mode == "hint" and context is not None:
                    context.append(
                        UserMessage(
                            content="Relevant plan:\n " + most_relevant_plan,
                            source=source,
                        )
                    )
        except Exception as e:
            trace_logger.error(f"Error retrieving plans from memory: {e}")
        return None

    async def _orchestrate_step_planning(
        self, cancellation_token: CancellationToken
    ) -> None:
        # Planning stage
        plan_response: Dict[str, Any] | None = None
        last_user_message = self._state.message_history[-1]
        assert last_user_message.source in [self._user_agent_topic, "user"]
        message_content: str = ""
        assert isinstance(last_user_message, TextMessage | MultiModalMessage)

        if isinstance(last_user_message.content, list):
            # iterate over the list and get the first item that is a string
            for item in last_user_message.content:
                if isinstance(item, str):
                    message_content = item
                    break
        else:
            message_content = last_user_message.content
        last_user_message = HumanInputFormat.from_str(message_content)

        # Is this our first time planning?
        if self._state.task == "" and self._state.plan_str == "":
            self._state.task = "TASK: " + last_user_message.content

            # TCM reuse plan
            from_memory = False
            if (
                not self._config.plan
                and self._config.retrieve_relevant_plans == "reuse"
            ):
                most_relevant_plan = await self._handle_relevant_plan_from_memory()
                if most_relevant_plan is not None:
                    self._config.plan = Plan.from_list_of_dicts_or_str(
                        most_relevant_plan
                    )
                    from_memory = True
            # Do we already have a plan to follow and planning mode is disabled?
            if self._config.plan is not None:
                self._state.plan = self._config.plan
                self._state.plan_str = str(self._config.plan)
                self._state.message_history.append(
                    TextMessage(
                        content="Initial plan from user:\n " + str(self._config.plan),
                        source="user",
                    )
                )
                plan_response = {
                    "task": self._state.plan.task,
                    "steps": [step.model_dump() for step in self._state.plan.steps],
                    "needs_plan": True,
                    "response": "",
                    "plan_summary": self._state.plan.task,
                    "from_memory": from_memory,
                }

                await self._log_message_agentchat(
                    dict_to_str(plan_response),
                    metadata={"internal": "no", "type": "plan_message"},
                )

                if not self._config.cooperative_planning:
                    self._state.in_planning_mode = False
                    await self._orchestrate_first_step(cancellation_token)
                    return
                else:
                    await self._request_next_speaker(
                        self._user_agent_topic, cancellation_token
                    )
                    return
            # Did the user provide a plan?
            user_plan = last_user_message.plan
            if user_plan is not None:
                self._state.plan = user_plan
                self._state.plan_str = str(user_plan)
                if last_user_message.accepted or is_accepted_str(
                    last_user_message.content
                ):
                    self._state.in_planning_mode = False
                    await self._orchestrate_first_step(cancellation_token)
                    return

            # assume the task is the last user message
            context = self._thread_to_context()
            # if bing search is enabled, do a bing search to help with planning
            if self._config.do_bing_search:
                bing_search_results = await self.do_bing_search(
                    last_user_message.content
                )
                if bing_search_results is not None:
                    context.append(
                        UserMessage(
                            content=bing_search_results,
                            source="web_surfer",
                        )
                    )

            if self._config.retrieve_relevant_plans == "hint":
                await self._handle_relevant_plan_from_memory(context=context)

            # create a first plan
            context.append(
                UserMessage(
                    content=self._get_task_ledger_plan_prompt(self._team_description),
                    source=self._name,
                )
            )
            plan_response = await self._get_json_response(
                context, self._validate_plan_json, cancellation_token
            )
            if self._state.is_paused:
                # let user speak next if paused
                await self._request_next_speaker(
                    self._user_agent_topic, cancellation_token
                )
                return
            assert plan_response is not None
            self._state.plan = Plan.from_list_of_dicts_or_str(plan_response["steps"])
            self._state.plan_str = str(self._state.plan)
            # add plan_response to the message thread
            self._state.message_history.append(
                TextMessage(
                    content=json.dumps(plan_response, indent=4), source=self._name
                )
            )
        else:
            # what did the user say?
            # Check if user accepted the plan
            if last_user_message.accepted or is_accepted_str(last_user_message.content):
                user_plan = last_user_message.plan
                if user_plan is not None:
                    self._state.plan = user_plan
                    self._state.plan_str = str(user_plan)
                # switch to execution mode
                self._state.in_planning_mode = False
                await self._orchestrate_first_step(cancellation_token)
                return
            # user did not accept the plan yet
            else:
                # update the plan
                user_plan = last_user_message.plan
                if user_plan is not None:
                    self._state.plan = user_plan
                    self._state.plan_str = str(user_plan)

                context = self._thread_to_context()

                # if bing search is enabled, do a bing search to help with planning
                if self._config.do_bing_search:
                    bing_search_results = await self.do_bing_search(
                        last_user_message.content
                    )
                    if bing_search_results is not None:
                        context.append(
                            UserMessage(
                                content=bing_search_results,
                                source="web_surfer",
                            )
                        )
                if self._config.retrieve_relevant_plans == "hint":
                    await self._handle_relevant_plan_from_memory(context=context)
                context.append(
                    UserMessage(
                        content=self._get_task_ledger_plan_prompt(
                            self._team_description
                        ),
                        source=self._name,
                    )
                )
                plan_response = await self._get_json_response(
                    context, self._validate_plan_json, cancellation_token
                )
                if self._state.is_paused:
                    # let user speak next if paused
                    await self._request_next_speaker(
                        self._user_agent_topic, cancellation_token
                    )
                    return
                assert plan_response is not None
                self._state.plan = Plan.from_list_of_dicts_or_str(
                    plan_response["steps"]
                )
                self._state.plan_str = str(self._state.plan)
                if not self._config.no_overwrite_of_task:
                    self._state.task = plan_response["task"]
                # add plan_response to the message thread
                self._state.message_history.append(
                    TextMessage(
                        content=json.dumps(plan_response, indent=4), source=self._name
                    )
                )

        assert plan_response is not None
        # if we don't need to plan, just send the message
        if self._config.cooperative_planning:
            if not plan_response["needs_plan"]:
                # send the response plan_message["response"] to the group
                await self._publish_group_chat_message(
                    plan_response["response"], cancellation_token
                )
                await self._request_next_speaker(
                    self._user_agent_topic, cancellation_token
                )
                return
            else:
                await self._publish_group_chat_message(
                    dict_to_str(plan_response),
                    cancellation_token,
                    metadata={"internal": "no", "type": "plan_message"},
                )
                await self._request_next_speaker(
                    self._user_agent_topic, cancellation_token
                )
                return
        else:
            await self._publish_group_chat_message(
                dict_to_str(plan_response),
                metadata={"internal": "no", "type": "plan_message"},
                cancellation_token=cancellation_token,
            )
            self._state.in_planning_mode = False
            await self._orchestrate_first_step(cancellation_token)

    async def _orchestrate_first_step(
        self, cancellation_token: CancellationToken
    ) -> None:
        # remove all messages from the message thread that are not from the user
        self._state.message_history = [
            m
            for m in self._state.message_history
            if m.source not in ["user", self._user_agent_topic]
        ]

        ledger_message = TextMessage(
            content=self._get_task_ledger_full_prompt(
                self._state.task, self._team_description, self._state.plan_str
            ),
            source=self._name,
        )
        # add the ledger message to the message thread internally
        self._state.message_history.append(ledger_message)
        await self._log_message_agentchat(ledger_message.content, internal=True)

        # check if the plan is empty, complete, or we have reached the max turns
        if (
            (not isinstance(self._state.plan, Plan) or len(self._state.plan) == 0)
            or (self._state.current_step_idx >= len(self._state.plan))
            or (
                self._config.max_turns is not None
                and self._state.n_rounds > self._config.max_turns
            )
        ):
            await self._prepare_final_answer("Max rounds reached.", cancellation_token)
            return
        self._state.n_rounds += 1
        context = self._thread_to_context()
        # Creat the progress ledger

        progress_ledger_prompt = self._get_progress_ledger_prompt(
            self._state.task,
            self._state.plan_str,
            self._state.current_step_idx,
            self._team_description,
            self._agent_execution_names,
        )
        context.append(UserMessage(content=progress_ledger_prompt, source=self._name))

        progress_ledger = await self._get_json_response(
            context, self._validate_ledger_json, cancellation_token
        )
        if self._state.is_paused:
            # let user speak next if paused
            await self._request_next_speaker(self._user_agent_topic, cancellation_token)
            return
        assert progress_ledger is not None

        await self._log_message_agentchat(dict_to_str(progress_ledger), internal=True)

        # Broadcast the next step
        new_instruction = self.get_agent_instruction(
            progress_ledger["instruction_or_question"]["answer"],
            progress_ledger["instruction_or_question"]["agent_name"],
        )

        message_to_send = TextMessage(
            content=new_instruction, source=self._name, metadata={"internal": "yes"}
        )
        self._state.message_history.append(message_to_send)  # My copy

        await self._publish_group_chat_message(
            message_to_send.content, cancellation_token, internal=True
        )
        json_step_execution = {
            "title": self._state.plan[self._state.current_step_idx].title,
            "index": self._state.current_step_idx,
            "details": self._state.plan[self._state.current_step_idx].details,
            "agent_name": progress_ledger["instruction_or_question"]["agent_name"],
            "instruction": progress_ledger["instruction_or_question"]["answer"],
            "progress_summary": progress_ledger["progress_summary"],
            "plan_length": len(self._state.plan),
        }
        await self._log_message_agentchat(
            json.dumps(json_step_execution),
            metadata={"internal": "no", "type": "step_execution"},
        )
        # Request that the step be completed
        valid_next_speaker: bool = False
        next_speaker = progress_ledger["instruction_or_question"]["agent_name"]
        for participant_name in self._agent_execution_names:
            if participant_name == next_speaker:
                await self._request_next_speaker(next_speaker, cancellation_token)
                valid_next_speaker = True
                break
        if not valid_next_speaker:
            raise ValueError(
                f"Invalid next speaker: {next_speaker} from the ledger, participants are: {self._agent_execution_names}"
            )

    async def _orchestrate_step_execution(
        self, cancellation_token: CancellationToken
    ) -> None:
        # Execution stage
        # Check if we reached the maximum number of rounds
        assert self._state.plan is not None
        if self._state.current_step_idx >= len(self._state.plan) or (
            self._config.max_turns is not None
            and self._state.n_rounds > self._config.max_turns
        ):
            await self._prepare_final_answer("Max rounds reached.", cancellation_token)
            return

        self._state.n_rounds += 1
        context = self._thread_to_context()
        # Update the progress ledger

        progress_ledger_prompt = self._get_progress_ledger_prompt(
            self._state.task,
            self._state.plan_str,
            self._state.current_step_idx,
            self._team_description,
            self._agent_execution_names,
        )

        context.append(UserMessage(content=progress_ledger_prompt, source=self._name))

        progress_ledger = await self._get_json_response(
            context, self._validate_ledger_json, cancellation_token
        )
        if self._state.is_paused:
            await self._request_next_speaker(self._user_agent_topic, cancellation_token)
            return
        assert progress_ledger is not None
        # log the progress ledger
        await self._log_message_agentchat(dict_to_str(progress_ledger), internal=True)

        # Check for replans
        need_to_replan = progress_ledger["need_to_replan"]["answer"]
        replan_reason = progress_ledger["need_to_replan"]["reason"]

        if need_to_replan and self._config.allow_for_replans:
            # Replan
            if self._config.max_replans is None:
                await self._replan(replan_reason, cancellation_token)
            elif self._state.n_replans < self._config.max_replans:
                self._state.n_replans += 1
                await self._replan(replan_reason, cancellation_token)
                return
            else:
                await self._prepare_final_answer(
                    f"We need to replan but max replan attempts reached: {replan_reason}.",
                    cancellation_token,
                )
                return
        elif need_to_replan:
            await self._prepare_final_answer(
                f"The current plan failed to complete the task, we need a new plan to continue. {replan_reason}",
                cancellation_token,
            )
            return
        if progress_ledger["is_current_step_complete"]["answer"]:
            self._state.current_step_idx += 1

        if progress_ledger["progress_summary"] != "":
            self._state.information_collected += (
                "\n" + progress_ledger["progress_summary"]
            )
        # Check for plan completion
        if self._state.current_step_idx >= len(self._state.plan):
            await self._prepare_final_answer(
                "Plan completed.",
                cancellation_token,
            )
            return

        # Broadcast the next step
        new_instruction = self.get_agent_instruction(
            progress_ledger["instruction_or_question"]["answer"],
            progress_ledger["instruction_or_question"]["agent_name"],
        )
        message_to_send = TextMessage(
            content=new_instruction, source=self._name, metadata={"internal": "yes"}
        )
        self._state.message_history.append(message_to_send)  # My copy

        await self._publish_group_chat_message(
            message_to_send.content, cancellation_token, internal=True
        )
        json_step_execution = {
            "title": self._state.plan[self._state.current_step_idx].title,
            "index": self._state.current_step_idx,
            "details": self._state.plan[self._state.current_step_idx].details,
            "agent_name": progress_ledger["instruction_or_question"]["agent_name"],
            "instruction": progress_ledger["instruction_or_question"]["answer"],
            "progress_summary": progress_ledger["progress_summary"],
            "plan_length": len(self._state.plan),
        }
        await self._log_message_agentchat(
            json.dumps(json_step_execution),
            metadata={"internal": "no", "type": "step_execution"},
        )

        # Request that the step be completed
        valid_next_speaker: bool = False
        next_speaker = progress_ledger["instruction_or_question"]["agent_name"]
        for participant_name in self._agent_execution_names:
            if participant_name == next_speaker:
                await self._request_next_speaker(next_speaker, cancellation_token)
                valid_next_speaker = True
                break
        if not valid_next_speaker:
            raise ValueError(
                f"Invalid next speaker: {next_speaker} from the ledger, participants are: {self._agent_execution_names}"
            )

    async def _replan(self, reason: str, cancellation_token: CancellationToken) -> None:
        # Let's create a new plan
        self._state.in_planning_mode = True
        await self._log_message_agentchat(
            f"We need to create a new plan. {reason}",
            metadata={"internal": "no", "type": "replanning"},
        )
        context = self._thread_to_context()

        # Store completed steps
        completed_steps = (
            list(self._state.plan.steps[: self._state.current_step_idx])
            if self._state.plan
            else []
        )
        completed_plan_str = "\n".join(
            [f"COMPLETED STEP {i+1}: {step}" for i, step in enumerate(completed_steps)]
        )

        # Add completed steps info to replan prompt
        replan_prompt = self._get_task_ledger_replan_plan_prompt(
            self._state.task,
            self._team_description,
            f"Completed steps so far:\n{completed_plan_str}\n\nPrevious plan:\n{self._state.plan_str}",
        )
        context.append(
            UserMessage(
                content=replan_prompt,
                source=self._name,
            )
        )
        plan_response = await self._get_json_response(
            context, self._validate_plan_json, cancellation_token
        )
        assert plan_response is not None

        # Create new plan by combining completed steps with new steps
        new_plan = Plan.from_list_of_dicts_or_str(plan_response["steps"])
        if new_plan is not None:
            combined_steps = completed_steps + list(new_plan.steps)
            self._state.plan = Plan(task=self._state.task, steps=combined_steps)
            self._state.plan_str = str(self._state.plan)
        else:
            # If new plan is None, keep the completed steps
            self._state.plan = Plan(task=self._state.task, steps=completed_steps)
            self._state.plan_str = str(self._state.plan)

        # Update task if in planning mode
        if not self._config.no_overwrite_of_task:
            self._state.task = plan_response["task"]

        plan_response["plan_summary"] = "Replanning: " + plan_response["plan_summary"]
        # Log the plan response in the same format as planning mode
        await self._publish_group_chat_message(
            dict_to_str(plan_response),
            cancellation_token=cancellation_token,
            metadata={"internal": "no", "type": "plan_message"},
        )
        # next speaker is user
        if self._config.cooperative_planning:
            await self._request_next_speaker(self._user_agent_topic, cancellation_token)
        else:
            self._state.in_planning_mode = False
            await self._orchestrate_first_step(cancellation_token)

    async def _prepare_final_answer(
        self,
        reason: str,
        cancellation_token: CancellationToken,
        final_answer: str | None = None,
        force_stop: bool = False,
    ) -> None:
        """Prepare the final answer for the task.

        Args:
            reason (str): The reason for preparing the final answer
            cancellation_token (CancellationToken): Token for cancellation
            final_answer (str, optional): Optional pre-computed final answer to use instead of computing one
            force_stop (bool): Whether to force stop the conversation after the final answer is computed
        """
        if final_answer is None:
            context = self._thread_to_context()
            # add replan reason
            context.append(UserMessage(content=reason, source=self._name))
            # Get the final answer
            final_answer_prompt = self._get_final_answer_prompt(self._state.task)
            progress_summary = f"Progress Summary:\n{self._state.information_collected}"
            context.append(
                UserMessage(
                    content=progress_summary + "\n\n" + final_answer_prompt,
                    source=self._name,
                )
            )

            # Re-initialize model context to meet token limit quota
            await self._model_context.clear()
            for msg in context:
                await self._model_context.add_message(msg)
            token_limited_context = await self._model_context.get_messages()

            response = await self._model_client.create(
                token_limited_context, cancellation_token=cancellation_token
            )
            assert isinstance(response.content, str)
            final_answer = response.content

        message = TextMessage(
            content=f"Final Answer: {final_answer}", source=self._name
        )

        self._state.message_history.append(message)  # My copy

        await self._publish_group_chat_message(
            message.content,
            cancellation_token,
            metadata={"internal": "no", "type": "final_answer"},
        )

        # reset internals except message history
        self._state.reset_for_followup()
        if not force_stop and self._config.allow_follow_up_input:
            await self._request_next_speaker(self._user_agent_topic, cancellation_token)
        else:
            # Signal termination
            await self._signal_termination(
                StopMessage(content=reason, source=self._name)
            )

        if self._termination_condition is not None:
            await self._termination_condition.reset()

    def _thread_to_context(
        self, messages: Optional[List[BaseChatMessage | BaseAgentEvent]] = None
    ) -> List[LLMMessage]:
        """Convert the message thread to a context for the model."""
        chat_messages: List[BaseChatMessage | BaseAgentEvent] = (
            messages if messages is not None else self._state.message_history
        )
        context_messages: List[LLMMessage] = []
        date_today = datetime.now().strftime("%d %B, %Y")
        if self._state.in_planning_mode:
            context_messages.append(
                SystemMessage(content=self._get_system_message_planning())
            )
        else:
            context_messages.append(
                SystemMessage(
                    content=ORCHESTRATOR_SYSTEM_MESSAGE_EXECUTION.format(
                        date_today=date_today
                    )
                )
            )
        if self._model_client.model_info["vision"]:
            context_messages.extend(
                thread_to_context(
                    messages=chat_messages, agent_name=self._name, is_multimodal=True
                )
            )
        else:
            context_messages.extend(
                thread_to_context(
                    messages=chat_messages, agent_name=self._name, is_multimodal=False
                )
            )
        return context_messages

    async def save_state(self) -> Mapping[str, Any]:
        """Save the current state of the orchestrator.

        Returns:
            Mapping[str, Any]: A dictionary containing all state attributes except is_paused.
        """
        # Get all state attributes except message_history and is_paused
        data = self._state.model_dump(exclude={"is_paused"})

        # Serialize message history separately to preserve type information
        data["message_history"] = [
            message.dump() for message in self._state.message_history
        ]

        # Serialize plan if it exists
        if self._state.plan is not None:
            data["plan"] = self._state.plan.model_dump()

        return data

    async def load_state(self, state: Mapping[str, Any]) -> None:
        """Load a previously saved state into the orchestrator.

        Args:
            state (Mapping[str, Any]): A dictionary containing the state attributes to load.
        """
        # Create new state with defaults
        new_state = OrchestratorState()

        # Load basic attributes
        for key, value in state.items():
            if key == "message_history":
                # Handle message history separately
                new_state.message_history = [
                    self._message_factory.create(message) for message in value
                ]
            elif key == "plan" and value is not None:
                # Reconstruct Plan object if it exists
                new_state.plan = Plan(**value)
            elif key != "is_paused" and hasattr(new_state, key):
                setattr(new_state, key, value)

        # Update the state
        self._state = new_state



from .web_surfer import WebSurfer, WebSurferCUA
from ._coder import CoderAgent
from ._user_proxy import USER_PROXY_DESCRIPTION
from .file_surfer import FileSurfer
from .drsai_agents import MagenticAgent
__all__ = [
    "WebSurfer",
    "WebSurferCUA",
    "CoderAgent",
    "USER_PROXY_DESCRIPTION",
    "FileSurfer",
    "MagenticAgent",
]



import inspect
from typing import Callable, List, Tuple


def _validate_path_impl() -> Callable[[str], None]:
    """The actual implementation of path validation"""
    import os

    def validate(path: str) -> None:
        if path == ".":
            print("true")
        else:
            print(os.path.exists(path))

    return validate


def _check_is_dir_impl() -> Callable[[str], None]:
    """The actual implementation of directory checking"""
    import os

    def check(path: str) -> None:
        if path == ".":
            print("true")
            return
        else:
            print(os.path.isdir(path))

    return check


def _convert_file_impl() -> Callable[[str], None]:
    """The actual implementation of file conversion"""
    from markitdown import MarkItDown

    def convert(path: str) -> None:
        converter = MarkItDown()
        result = converter.convert_local(path)
        print("TITLE:" + (result.title or ""))
        print("CONTENT:" + result.text_content)

    return convert


def _directory_listing_impl() -> Callable[[str], None]:
    """The actual implementation of directory listing"""
    import os
    import datetime

    def list_directory(path: str) -> None:
        listing = """
| Name | Size | Date Modified |
| ---- | ---- | ------------- |
| .. (parent directory) | | |
"""
        for entry in os.listdir(path):
            size = ""
            full_path = os.path.join(path, entry)

            mtime = ""
            try:
                mtime = datetime.datetime.fromtimestamp(
                    os.path.getmtime(full_path)
                ).strftime("%Y-%m-%d %H:%M")
            except Exception as e:
                mtime = f"N/A: {type(e).__name__}"

            if os.path.isdir(full_path):
                entry = entry + os.path.sep
            else:
                try:
                    size = str(os.path.getsize(full_path))
                except Exception as e:
                    size = f"N/A: {type(e).__name__}"

            listing += f"| {entry} | {size} | {mtime} |\n"
        print(listing)

    return list_directory


def _find_files_impl() -> Callable[[str], None]:
    """The actual implementation of file finding"""
    import os
    from difflib import SequenceMatcher

    def find_files(query: str) -> None:
        import json

        ROOT_DIR = "."
        THRESHOLD = 0.2
        MAX_RESULTS = 20
        INCLUDE_DIRECTORIES = False

        matches: List[Tuple[str, float]] = []
        perfect_match = None

        for root, dirs, files in os.walk(ROOT_DIR):
            # Add both files and directories
            dirs[:] = [
                d for d in dirs if d not in ["node_modules", ".git", "__pycache__"]
            ]
            if INCLUDE_DIRECTORIES:
                items = files + dirs
            else:
                items = files
            for name in items:
                path = os.path.join(root, name)[2:]  # Remove ./ prefix

                # Check for exact matches first (case insensitive)
                if name.lower() == query.lower():
                    score = 1.0
                    perfect_match = path
                else:
                    # Calculate similarity score
                    score = SequenceMatcher(None, query.lower(), name.lower()).ratio()

                if score > THRESHOLD:  # Minimum similarity threshold
                    matches.append((path, score))

        # Sort by score and take top results
        matches.sort(key=lambda x: x[1], reverse=True)
        matches = matches[:MAX_RESULTS]

        result = {"matches": matches, "perfect_match": perfect_match}

        print(json.dumps(result))

    return find_files


# The template functions that generate the code to be executed
def get_path_validation_code(path: str) -> str:
    inner_func = inspect.getsource(_validate_path_impl()).split("def validate", 1)[1]
    return f"""
import os
def validate{inner_func}
validate('{path}')
"""


def get_is_dir_check_code(path: str) -> str:
    inner_func = inspect.getsource(_check_is_dir_impl()).split("def check", 1)[1]
    return f"""
import os
def check{inner_func}
check('{path}')
"""


def get_file_conversion_code(path: str) -> str:
    inner_func = inspect.getsource(_convert_file_impl()).split("def convert", 1)[1]
    return f"""
from markitdown import MarkItDown
def convert{inner_func}
convert('{path}')
"""


def get_directory_listing_code(local_path: str) -> str:
    inner_func = inspect.getsource(_directory_listing_impl()).split(
        "def list_directory", 1
    )[1]
    return f"""
import os
import datetime
def list_directory{inner_func}
list_directory('{local_path}')
"""


def get_find_files_code(query: str) -> str:
    inner_func = inspect.getsource(_find_files_impl()).split("def find_files", 1)[1]
    return f"""
import os
from difflib import SequenceMatcher
def find_files{inner_func}
find_files('{query}')
"""



import json
import logging
import os
import psutil
from typing import List, Union, Dict

from autogen_core.models import (
    LLMMessage,
    UserMessage,
    AssistantMessage,
)

from autogen_agentchat.utils import remove_images
from autogen_agentchat.messages import (
    BaseChatMessage,
    HandoffMessage,
    MultiModalMessage,
    StopMessage,
    TextMessage,
    ToolCallRequestEvent,
    ToolCallExecutionEvent,
    BaseAgentEvent,
)

from .types import HumanInputFormat, RunPaths


class LLMCallFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        try:
            message = json.loads(record.getMessage())
            return message.get("type") == "LLMCall"
        except (json.JSONDecodeError, AttributeError):
            return False


# Define recursive types for JSON structures
JsonPrimitive = Union[str, int, float, bool, None]
JsonList = List[Union[JsonPrimitive, "JsonDict", "JsonList"]]
JsonDict = Dict[str, Union[JsonPrimitive, JsonList, "JsonDict"]]
JsonData = Union[JsonDict, JsonList, str]


def json_data_to_markdown(data: JsonData) -> str:
    """
    Convert a dictionary, list, or JSON string to a nicely formatted Markdown string.
    Handles nested structures of dictionaries and lists.

    Args:
        data (JsonData): The data to convert, can be:
            - A dictionary with string keys and JSON-compatible values
            - A list of JSON-compatible values
            - A JSON string representing either of the above

    Returns:
        str: The formatted Markdown string.

    Raises:
        ValueError: If the input cannot be parsed or converted to markdown format.
        json.JSONDecodeError: If the input string is not valid JSON.
    """

    def format_dict(d: JsonDict, indent: int = 0) -> str:
        md = ""
        for key, value in d.items():
            md += "  " * indent + f"- {key}: "
            if isinstance(value, dict):
                md += "\n" + format_dict(value, indent + 1)
            elif isinstance(value, list):
                md += "\n" + format_list(value, indent + 1)
            else:
                md += f"{value}\n"
        return md

    def format_list(lst: JsonList, indent: int = 0) -> str:
        md = ""
        for item in lst:
            if isinstance(item, dict):
                md += "  " * indent + "- \n" + format_dict(item, indent + 1)
            elif isinstance(item, list):
                md += "  " * indent + "- \n" + format_list(item, indent + 1)
            else:
                md += "  " * indent + f"- {item}\n"
        return md

    try:
        if isinstance(data, str):
            data = json.loads(data)

        if isinstance(data, list):
            return format_list(data)
        elif isinstance(data, dict):
            return format_dict(data)
        else:
            raise ValueError(f"Expected dict, list or JSON string, got {type(data)}")

    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(f"Invalid JSON string: {str(e)}", e.doc, e.pos)
    except Exception as e:
        raise ValueError(f"Failed to convert to markdown: {str(e)}")


def dict_to_str(data: Union[JsonDict, str]) -> str:
    """
    Convert a dictionary or JSON string to a JSON string.

    Args:
        data (JsonDict | str): The dictionary or JSON string to convert.

    Returns:
        str: The input dictionary in JSON format.
    """
    if isinstance(data, dict):
        return json.dumps(data)
    elif isinstance(data, str):
        return data
    else:
        raise ValueError("Unexpected input type")


def thread_to_context(
    messages: List[BaseAgentEvent | BaseChatMessage],
    agent_name: str,
    is_multimodal: bool = False,
) -> List[LLMMessage]:
    """Convert the message thread to a context for the model."""
    context: List[LLMMessage] = []
    for m in messages:
        if isinstance(m, ToolCallRequestEvent | ToolCallExecutionEvent):
            # Ignore tool call messages.
            continue
        elif isinstance(m, StopMessage | HandoffMessage):
            context.append(UserMessage(content=m.content, source=m.source))
        elif m.source == agent_name:
            assert isinstance(m, TextMessage)
            context.append(AssistantMessage(content=m.content, source=m.source))
        elif m.source == "user_proxy" or m.source == "user":
            assert isinstance(m, TextMessage | MultiModalMessage)
            if isinstance(m.content, str):
                human_input = HumanInputFormat.from_str(m.content)
                content = f"{human_input.content}"
                if human_input.plan is not None:
                    content += f"\n\nI created the following plan: {human_input.plan}"
                context.append(UserMessage(content=content, source=m.source))
            else:
                # If content is a list, transform only the string part
                content_list = list(m.content)  # Create a copy of the list
                for i, item in enumerate(content_list):
                    if isinstance(item, str):
                        human_input = HumanInputFormat.from_str(item)
                        content_list[i] = f"{human_input.content}"
                        if human_input.plan is not None and isinstance(
                            content_list[i], str
                        ):
                            content_list[i] = (
                                f"{content_list[i]}\n\nI created the following plan: {human_input.plan}"
                            )
                context.append(UserMessage(content=content_list, source=m.source))  # type: ignore
        else:
            assert isinstance(m, TextMessage) or isinstance(m, MultiModalMessage)
            context.append(UserMessage(content=m.content, source=m.source))
    if is_multimodal:
        return context
    else:
        return remove_images(context)


def get_internal_urls(inside_docker: bool, paths: RunPaths) -> List[str] | None:
    if not inside_docker:
        return None
    urls: List[str] = []
    for _, addrs in psutil.net_if_addrs().items():
        for addr in addrs:
            if addr.family.name == "AF_INET":
                urls.append(addr.address)

    hostname = os.getenv("HOSTNAME")
    if hostname is not None:
        urls.append(hostname)
    container_name = os.getenv("CONTAINER_NAME")
    if container_name is not None:
        urls.append(container_name)
    return urls



import json
from drsai import AssistantAgent, HepAIChatCompletionClient
import asyncio
from typing import AsyncGenerator,Optional, List, Dict, Tuple, Sequence, Any, Awaitable, Callable, Union
from loguru import logger
# from datetime import datetime
from pydantic import BaseModel


from autogen_core import CancellationToken, FunctionCall
from autogen_core.model_context import (
    ChatCompletionContext,
)
from autogen_agentchat.base import Response, TaskResult
from autogen_agentchat.base import Handoff as HandoffBase
from autogen_agentchat.messages import (
    StructuredMessageFactory,
    BaseChatMessage,
    TextMessage,
    HandoffMessage,
    StopMessage,
    ToolCallSummaryMessage,
    StructuredMessage,
    BaseAgentEvent,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    CodeGenerationEvent,
    CodeExecutionEvent,
    UserInputRequestedEvent,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    ThoughtEvent,
    SelectSpeakerEvent,
    SelectorEvent,
    MessageFactory,
    MultiModalMessage,
    Image,
)
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    CreateResult,
    RequestUsage,
    FunctionExecutionResult,
    FunctionExecutionResultMessage,
    LLMMessage,
    # ModelFamily,
    SystemMessage,
)
from autogen_core.tools import (
    BaseTool, 
    # FunctionTool, 
    # StaticWorkbench, 
    Workbench, 
    # ToolResult, 
    # TextResultContent, 
    # ToolSchema
    )

from ..magentic_one.guarded_action import ApprovalDeniedError
from hepai.tools.get_woker_functions import get_worker_sync_functions
from openai import Stream

class StatusAgent(AssistantAgent):
    '''
    连接OpenAI格式的模型或者智能体后端
    '''
    def __init__(
            self, 
            name: str,
            model_client: HepAIChatCompletionClient|None = None,
            model_client_stream: bool = True,
            tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
            description: str = "An agent that provides assistance with ability to use tools.",
            system_message: (
                str | None
            ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
            memory_function: Callable | None = None,
            allow_reply_function: bool = False,
            reply_function: Callable | None = None,
            model_remote_configs: Dict[str, Any] = {},
            chat_id: str|None = None,
            run_info: Dict[str, Any] = {},
            **kwargs):
        
        super().__init__(
            name = name, 
            model_client = model_client,
            model_client_stream = model_client_stream,
            tools = tools,
            description = description,
            system_message = system_message,
            memory_function = memory_function,
            allow_reply_function = allow_reply_function,
            reply_function = reply_function,
            **kwargs)

        self.is_paused = False
        self._paused = asyncio.Event()

        self._chat_id = chat_id
        run_info.update(kwargs)
        self._run_info = run_info 
        
        # initialize the sync model client
        self.api_key = model_remote_configs.pop("api_key", "")
        self.url = model_remote_configs.pop("url", "https://aiapi.ihep.ac.cn/apiv2")
        self.model_name = model_remote_configs.pop("model", "hepai/drsai")

        # worker函数
        self._funcs_map = {}

        self._init_message: str|dict = ""

        # 消息类型
        self._message_factory = MessageFactory()

    async def lazy_init(self, **kwargs) -> None:
        """Initialize the tools and models needed by the agent."""
        try:
            funcs = get_worker_sync_functions(
                name=self.model_name, 
                api_key=self.api_key,
                base_url=self.url,
            )
            # print([f.__name__ for f in funcs])
            self._funcs_map = {f.__name__: f for f in funcs}
            result: Dict[str, Any] = self._funcs_map['lazy_init'](chat_id=self._chat_id, api_key=self.api_key)
            status = result.get("status", False)
            message = result.get("message", "")
            if message:
                self._init_message = message
            if not status:
                raise Exception(message)
            else:
                logger.info(f"Lazy init {self.name} successfully.")
        except Exception as e:
            logger.error(f"Failed to load worker functions: {e}")
            self._funcs_map = {}
        

    async def pause(self) -> None:
        """Pause the agent by setting the paused state."""
        logger.info(f"Paused {self.name}...")

        # 停掉远程的模型
        result: Dict[str, Any] = self._funcs_map['pause'](chat_id=self._chat_id)
        status = result.get("status", False)
        message = result.get("message", "")
        if not status:
            raise Exception(message)
        else:
            logger.info(f"Paused {self.name} successfully.")

        # 设置暂停状态
        self.is_paused = True
        self._paused.set()
    
    async def pause_long_task(self) -> None:
        """Pause the long task by setting the paused state."""

        # 停掉远程的long task
        result: Dict[str, Any] = self._funcs_map['pause_long_task'](chat_id=self._chat_id)
        status = result.get("status", False)
        message = result.get("message", "")
        if not status:
            raise Exception(message)
        else:
            logger.info(f"Paused long task {self.name} successfully.")

    async def resume(self) -> None:
        """Resume the agent by clearing the paused state."""

        # 恢复远程的模型
        result: Dict[str, Any] = self._funcs_map['resume'](chat_id=self._chat_id)
        status = result.get("status", False)
        message = result.get("message", "")
        if not status:
            raise Exception(message)
        else:
            logger.info(f"Resumed {self.name} successfully.")

        # 取消暂停状态
        self.is_paused = False
        self._paused.clear()

    async def close(self) -> None:
        """Clean up resources used by the agent.

        This method:
          ...
        """
        logger.info(f"Closing {self.name}...")

        # 关闭模型客户端
        if self._model_client:
            await self._model_client.close()

        result: Dict[str, Any] = self._funcs_map['close'](chat_id=self._chat_id)
        status = result.get("status", False)
        message = result.get("message", "")
        if not status:
            raise Exception(message)
        else:
            logger.info(f"Closed {self.name} successfully.")

    async def async_stream_generator(self, stream, timeout: float = 180.0) -> AsyncGenerator[dict, None]:
        loop = asyncio.get_event_loop()
        queue = asyncio.Queue()

        def sync_consumer():
            try:
                for chunk in stream:
                    # # 检查暂停状态
                    # if self.is_paused:
                    #     loop.call_soon_threadsafe(queue.put_nowait, asyncio.CancelledError())
                    #     return
                    loop.call_soon_threadsafe(queue.put_nowait, chunk)
            except Exception as e:
                loop.call_soon_threadsafe(queue.put_nowait, e)
            finally:
                loop.call_soon_threadsafe(queue.put_nowait, None)  # 结束信号

        executor_task = loop.run_in_executor(None, sync_consumer)

        try:
            while True:
                item = await asyncio.wait_for(queue.get(), timeout=timeout)
                if item is None:  # 正常结束
                    break
                elif isinstance(item, type) and issubclass(item, Exception):
                    # 如果是异常类型，实例化并抛出
                    # if issubclass(item, asyncio.CancelledError):
                    #     raise asyncio.CancelledError("Stream cancelled")
                    # else:
                    #     raise item()
                    raise item()
                elif isinstance(item, Exception):
                    raise item
                else:
                    yield item
        except asyncio.TimeoutError:
            logger.warning(f"Stream timeout after {timeout} seconds")
            raise asyncio.TimeoutError("Model streaming timed out")
        finally:
            if not executor_task.done():
                executor_task.cancel()
            try:
                await executor_task
            except asyncio.CancelledError:
                pass

    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """Handle incoming messages and yield responses as a stream. Append the request to agents chat history."""

        if not self._funcs_map:
            yield Response(
                chat_message=TextMessage(
                    content=f"Cannot connect to the model: {self.model_name} through {self.url}. Please check the connection and try again later.",
                    source=self.name,
                    metadata={"internal": "No"},
                )
            )
            return
        
        # monitor the pause event
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content=f"The {self.name} is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return

        # Set up background task to monitor the pause event and cancel the task if paused.
        # async def monitor_pause() -> None:
        #     await self._paused.wait()
        #     self.is_paused = True
        # monitor_pause_task = asyncio.create_task(monitor_pause())

        
        try:
        ##########Your costum code here##########
        # NOTE: Can only yield TextMessage or MultiModalMessage in MagenticAgent becasue the limit in src/magentic_ui/utils.py", line 160, in thread_to_context:
        # ```assert isinstance(m, TextMessage) or isinstance(m, MultiModalMessage)```

            # Gather all relevant state here
            agent_name = self.name
            model_context = self._model_context
            memory = self._memory
            system_messages = self._system_messages
            workbench = self._workbench
            handoff_tools = self._handoff_tools
            handoffs = self._handoffs
            model_client = self._model_client
            model_client_stream = self._model_client_stream
            reflect_on_tool_use = self._reflect_on_tool_use
            tool_call_summary_format = self._tool_call_summary_format
            output_content_type = self._output_content_type
            format_string = self._output_content_type_format

            # STEP 1: Add new user/handoff messages to the model context
            ## 将前端传入的json格式的user message转换为str
            try:
                input = messages[-1].content
                data = json.loads(input)
                if not isinstance(data, dict):
                    raise ValueError("Input string must be a JSON object")
                input_str = data.get("content", "")
            except (json.JSONDecodeError, ValueError):
                input_str = messages[-1].content
            
            messages[-1].content = input_str
            
            await self._add_messages_to_context(
                model_context=model_context,
                messages=messages,
            )

            # STEP 2: Update model context with any relevant memory
            inner_messages: List[BaseAgentEvent | BaseChatMessage] = []
            for event_msg in await self._update_model_context_with_memory(
                memory=memory,
                model_context=model_context,
                agent_name=agent_name,
            ):
                inner_messages.append(event_msg)
                yield event_msg

        
            # STEP 3: Run the first inference
            model_result = None
            # all_messages = await model_context.get_messages()
            # # llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=system_messages + all_messages)
            # # not add system_messages to llm_messages
            # llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=all_messages)
            # oai_massages = await self.llm_messages2oai_messages(llm_messages)
            
            # NOTE: 请注意，这是一个同步的迭代器，会堵塞当前线程，直到模型返回结果

            stream: Stream = self._funcs_map['a_chat_completions'](
                messages = [message.model_dump(mode="json") for message in messages],
                apikey = self.api_key,
                stream=True,
                model = agent_name,
                chat_id = self._chat_id,
                user = self._run_info,
            )
            full_response = []
            try:
                async for chunk in self.async_stream_generator(stream):
                    if self.is_paused:
                        raise asyncio.CancelledError("Agent paused during streaming")
                    # textchunk = chunk["choices"][0]["delta"].get("content", "")
                    # if textchunk:
                    #     yield ModelClientStreamingChunkEvent(content=textchunk, source=agent_name)
                    #     full_response += textchunk
                    # else:
                    #     if chunk["choices"][0].get("finish_reason") == "stop":
                    #         break
                    message_type = chunk.get("type", None)
                    if message_type in self._message_factory._message_types:
                        msg: BaseChatMessage|BaseAgentEvent = self._message_factory._message_types[message_type].model_validate(chunk)
                        
                        if message_type in [
                            "TextMessage",
                            "ToolCallSummaryMessage",
                            "StructuredMessage",
                            "HandoffMessage",
                            "StopMessage"]:
                            full_response.append({msg.source: msg.content})
                        else:
                            yield msg
                    if "stop_reason" in chunk:
                        # taskresult = TaskResult.model_validate(chunk)
                        break
                    
            except asyncio.CancelledError:
                raise
            except Exception as e:
                logger.error(f"Error during streaming: {e}")
                raise
            # for chunk in stream:
            #     if self.is_paused:
            #         raise asyncio.CancelledError()
            #     textchunck = chunk["choices"][0]["delta"]["content"]
            #     if textchunck:
            #         yield ModelClientStreamingChunkEvent(content=textchunck, source=agent_name)
            #         full_response += textchunck
            #     else:
            #         if chunk["choices"][0]["finish_reason"]:
            #             if chunk["choices"][0]["finish_reason"] == "stop": 
            #                 break

            full_response_str = ""
            if len(full_response)>1:
                for response in full_response:
                    for key, value in response.items():
                        full_response_str += f"**{key}:**\n\n{value}\n\n"
            else:
                for response in full_response:
                    for key, value in response.items():
                        full_response_str += f"{value}"

            model_result = CreateResult(
                content=full_response_str, 
                finish_reason="stop",
                usage = RequestUsage(prompt_tokens = 0, completion_tokens = len(full_response_str.split())),
                cached = False
                )
            

            assert model_result is not None, "No model result was produced."

            # --- NEW: If the model produced a hidden "thought," yield it as an event ---
            if model_result.thought:
                thought_event = ThoughtEvent(content=model_result.thought, source=agent_name)
                yield thought_event
                inner_messages.append(thought_event)

            # Add the assistant message to the model context (including thought if present)
            await model_context.add_message(
                AssistantMessage(
                    content=model_result.content,
                    source=agent_name,
                    thought=getattr(model_result, "thought", None),
                )
            )

            # STEP 4: Process the model output
            async for output_event in self._process_model_result(
                model_result=model_result,
                inner_messages=inner_messages,
                cancellation_token=cancellation_token,
                agent_name=agent_name,
                system_messages=system_messages,
                model_context=model_context,
                workbench=workbench,
                handoff_tools=handoff_tools,
                handoffs=handoffs,
                model_client=model_client,
                model_client_stream=model_client_stream,
                reflect_on_tool_use=reflect_on_tool_use,
                tool_call_summary_format=tool_call_summary_format,
                output_content_type=output_content_type,
                format_string=format_string,
            ):
                yield output_event

        ##########Your costum code above##########

        except ApprovalDeniedError:
            # If the user denies the approval, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The user did not approve the task.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except asyncio.TimeoutError:
            # If the task times out, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task timed out.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except Exception as e:
            logger.error(f"Error in {self.name}: {e}")
            # add to chat history
            await model_context.add_message(
                AssistantMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content="An error occurred while executing the task.",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:

            # # Cancel the monitor task.
            # try:
            #     monitor_pause_task.cancel()
            #     await monitor_pause_task
            # except asyncio.CancelledError:
            #     pass
            pass
    
    @classmethod
    async def _process_model_result(
        cls,
        model_result: CreateResult,
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        cancellation_token: CancellationToken,
        agent_name: str,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        handoffs: Dict[str, HandoffBase],
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        reflect_on_tool_use: bool,
        tool_call_summary_format: str,
        output_content_type: type[BaseModel] | None,
        format_string: str | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        Handle final or partial responses from model_result, including tool calls, handoffs,
        and reflection if needed.

        NOTE: Can only yield TextMessage or MultiModalMessage in MagenticAgent becasue the limit in src/magentic_ui/utils.py", line 160, in thread_to_context:
        ```assert isinstance(m, TextMessage) or isinstance(m, MultiModalMessage)```

        """

        # If direct text response (string)
        if isinstance(model_result.content, str):
            # if output_content_type:
            #     content = output_content_type.model_validate_json(model_result.content)
            #     yield Response(
            #         chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
            #             content=content,
            #             source=agent_name,
            #             models_usage=model_result.usage,
            #             format_string=format_string,
            #         ),
            #         inner_messages=inner_messages,
            #     )
            # else:
            #     yield Response(
            #         chat_message=TextMessage(
            #             content=model_result.content,
            #             source=agent_name,
            #             models_usage=model_result.usage,
            #         ),
            #         inner_messages=inner_messages,
            #     )
            yield Response(
                chat_message=TextMessage(
                    content=model_result.content,
                    source=agent_name,
                    models_usage=model_result.usage,
                    metadata={"internal": "no"}, # detect if it is internal message or not
                ),
                inner_messages=inner_messages,
            )
            return

        # Otherwise, we have function calls
        assert isinstance(model_result.content, list) and all(
            isinstance(item, FunctionCall) for item in model_result.content
        )

        # cannot yield ToolCallRequestEvent in MagenticAgent
        # STEP 4A: Yield ToolCallRequestEvent
        tool_call_msg = ToolCallRequestEvent(
            content=model_result.content,
            source=agent_name,
            models_usage=model_result.usage,
        )

        # event_logger.debug(tool_call_msg)
        logger.info(f"{cls.name} handling tool call with params : {model_result.content}")
        inner_messages.append(tool_call_msg)
        yield tool_call_msg

        # STEP 4B: Execute tool calls
        executed_calls_and_results = await asyncio.gather(
            *[
                cls._execute_tool_call(
                    tool_call=call,
                    workbench=workbench,
                    handoff_tools=handoff_tools,
                    agent_name=agent_name,
                    cancellation_token=cancellation_token,
                )
                for call in model_result.content
            ]
        )

        # cannot yield ToolCallExecutionEvent in MagenticAgent
        exec_results = [result for _, result in executed_calls_and_results]

        # Yield ToolCallExecutionEvent
        tool_call_result_msg = ToolCallExecutionEvent(
            content=exec_results,
            source=agent_name,
        )
        # event_logger.debug(tool_call_result_msg)
        await model_context.add_message(FunctionExecutionResultMessage(content=exec_results))
        inner_messages.append(tool_call_result_msg)
        yield tool_call_result_msg

        # cannot yield HandoffMessage
        # # STEP 4C: Check for handoff
        # handoff_output = cls._check_and_handle_handoff(
        #     model_result=model_result,
        #     executed_calls_and_results=executed_calls_and_results,
        #     inner_messages=inner_messages,
        #     handoffs=handoffs,
        #     agent_name=agent_name,
        # )
        # if handoff_output:
        #     yield handoff_output
        #     return

        # STEP 4D: Reflect or summarize tool results
        if reflect_on_tool_use:
            async for reflection_response in cls._reflect_on_tool_use_flow(
                system_messages=system_messages,
                model_client=model_client,
                model_client_stream=model_client_stream,
                model_context=model_context,
                agent_name=agent_name,
                inner_messages=inner_messages,
                output_content_type=output_content_type,
            ):
                yield reflection_response
        else:
            yield cls._summarize_tool_use(
                executed_calls_and_results=executed_calls_and_results,
                inner_messages=inner_messages,
                handoffs=handoffs,
                tool_call_summary_format=tool_call_summary_format,
                agent_name=agent_name,
            )

    @classmethod
    async def _reflect_on_tool_use_flow(
        cls,
        system_messages: List[SystemMessage],
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        model_context: ChatCompletionContext,
        agent_name: str,
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[Response | ModelClientStreamingChunkEvent | ThoughtEvent, None]:
        """
        If reflect_on_tool_use=True, we do another inference based on tool results
        and yield the final text response (or streaming chunks).
        """
        all_messages = system_messages + await model_context.get_messages()
        llm_messages = cls._get_compatible_context(model_client=model_client, messages=all_messages)

        reflection_result: Optional[CreateResult] = None

        if model_client_stream:
            async for chunk in model_client.create_stream(
                llm_messages,
                json_output=output_content_type,
            ):
                if isinstance(chunk, CreateResult):
                    reflection_result = chunk
                elif isinstance(chunk, str):
                    yield ModelClientStreamingChunkEvent(content=chunk, source=agent_name)
                else:
                    raise RuntimeError(f"Invalid chunk type: {type(chunk)}")
        else:
            reflection_result = await model_client.create(llm_messages, json_output=output_content_type)

        if not reflection_result or not isinstance(reflection_result.content, str):
            raise RuntimeError("Reflect on tool use produced no valid text response.")

        # --- NEW: If the reflection produced a thought, yield it ---
        if reflection_result.thought:
            thought_event = ThoughtEvent(content=reflection_result.thought, source=agent_name)
            yield thought_event
            inner_messages.append(thought_event)

        # Add to context (including thought if present)
        await model_context.add_message(
            AssistantMessage(
                content=reflection_result.content,
                source=agent_name,
                thought=getattr(reflection_result, "thought", None),
            )
        )

        # if output_content_type:
        #     content = output_content_type.model_validate_json(reflection_result.content)
        #     yield Response(
        #         chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
        #             content=content,
        #             source=agent_name,
        #             models_usage=reflection_result.usage,
        #         ),
        #         inner_messages=inner_messages,
        #     )
        # else:
        #     yield Response(
        #         chat_message=TextMessage(
        #             content=reflection_result.content,
        #             source=agent_name,
        #             models_usage=reflection_result.usage,
        #         ),
        #         inner_messages=inner_messages,
        #     )

        yield Response(
            chat_message=TextMessage(
                content=reflection_result.content,
                source=agent_name,
                models_usage=reflection_result.usage,
                metadata={"internal": "no"}, # detect if it is internal message or not
            ),
            inner_messages=inner_messages,
        )

    @staticmethod
    def _summarize_tool_use(
        executed_calls_and_results: List[Tuple[FunctionCall, FunctionExecutionResult]],
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        handoffs: Dict[str, HandoffBase],
        tool_call_summary_format: str,
        agent_name: str,
    ) -> Response:
        """
        If reflect_on_tool_use=False, create a summary message of all tool calls.
        """
        # Filter out calls which were actually handoffs
        normal_tool_calls = [(call, result) for call, result in executed_calls_and_results if call.name not in handoffs]
        tool_call_summaries: List[str] = []
        for tool_call, tool_call_result in normal_tool_calls:
            # 对MCP的结果进行处理
            try:
                json_results = json.loads(tool_call_result.content)
                if isinstance(json_results, list):
                    json_result = json_results[0]
                    if isinstance(json_result, dict) and 'type' in json_result:
                        if json_result['type'] == 'text':
                            tool_call_result.content = json_result['text']
            except:
                pass
            tool_call_summaries.append(
                tool_call_summary_format.format(
                    tool_name=tool_call.name,
                    arguments=tool_call.arguments,
                    result=tool_call_result.content,
                )
            )
        tool_call_summary = "\n".join(tool_call_summaries)
        # return Response(
        #     chat_message=ToolCallSummaryMessage(
        #         content=tool_call_summary,
        #         source=agent_name,
        #     ),
        #     inner_messages=inner_messages,
        # )

        return Response(
            chat_message=TextMessage(
                content=tool_call_summary,
                source=agent_name,
                metadata={"internal": "no"}, # detect if it is internal message or not
            ),
            inner_messages=inner_messages,
        )
    
    async def run_stream(
        self,
        *,
        task: str | BaseChatMessage | Sequence[BaseChatMessage] | None = None,
        cancellation_token: CancellationToken | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | TaskResult, None]:
        """Run the agent with the given task and return a stream of messages
        and the final task result as the last item in the stream."""
        if cancellation_token is None:
            cancellation_token = CancellationToken()
        input_messages: List[BaseChatMessage] = []
        output_messages: List[BaseAgentEvent | BaseChatMessage] = []
        if task is None:
            pass
        elif isinstance(task, str):
            text_msg = TextMessage(content=task, source="user", metadata={"internal": "yes"})
            input_messages.append(text_msg)
            output_messages.append(text_msg)
            yield text_msg
        elif isinstance(task, BaseChatMessage):
            input_messages.append(task)
            output_messages.append(task)
            task.metadata["internal"] = "yes"
            yield task
        else:
            if not task:
                raise ValueError("Task list cannot be empty.")
            for msg in task:
                if isinstance(msg, BaseChatMessage):
                    input_messages.append(msg)
                    output_messages.append(msg)
                    msg.metadata["internal"] = "yes"
                    yield msg
                else:
                    raise ValueError(f"Invalid message type in sequence: {type(msg)}")
        async for message in self.on_messages_stream(input_messages, cancellation_token):
            if isinstance(message, Response):
                yield message.chat_message
                output_messages.append(message.chat_message)
                yield TaskResult(messages=output_messages)
                
            else:
                yield message
                if isinstance(message, ModelClientStreamingChunkEvent):
                    # Skip the model client streaming chunk events.
                    continue
                output_messages.append(message)




from typing import Protocol, Literal, Optional, cast
from autogen_core import CancellationToken

from inspect import iscoroutinefunction
from autogen_agentchat.agents._user_proxy_agent import InputFuncType as AGInputFuncType

# This is a workaround to avoid modifying the autogen submodule; ideally we want to promote
# some form of "multiple types of input" to autogen_agentchat, but we should bake it here
# first.

InputRequestType = Literal["text_input", "approval"]


class SyncInputFunc(Protocol):
    """Protocol for synchronous input function."""

    def __call__(self, prompt: str, input_type: InputRequestType = "text_input") -> str:
        """Call the input function with a prompt and optional cancellation token."""
        ...


class AsyncInputFunc(Protocol):
    """Protocol for asynchronous input function."""

    async def __call__(
        self,
        prompt: str,
        cancellation_token: Optional[CancellationToken],
        input_type: InputRequestType = "text_input",
    ) -> str:
        """Call the input function with a prompt and optional cancellation token."""
        ...


InputFuncType = SyncInputFunc | AsyncInputFunc


def make_agentchat_input_func(
    input_func: Optional[InputFuncType] = None,
) -> Optional[AGInputFuncType]:
    """Convert a custom input function to the AgentChat input function format."""
    if input_func is None:
        return None

    if iscoroutinefunction(input_func):
        actual_input_func = cast(AsyncInputFunc, input_func)

        async def async_input_func(
            prompt: str, cancellation_token: Optional[CancellationToken]
        ) -> str:
            return await actual_input_func(
                prompt, cancellation_token, input_type="text_input"
            )

        return async_input_func
    else:
        actual_input_func = cast(SyncInputFunc, input_func)

        def sync_input_func(prompt: str) -> str:
            return actual_input_func(prompt, input_type="text_input")

        return sync_input_func



import math
import traceback
from enum import Enum
from functools import wraps

from fastapi import HTTPException
from pydantic import BaseModel, ConfigDict

# from haiddf.components.portal.log import logger
from loguru import logger


class PortalErrCode(Enum):
    """
    只用于业务代码。不要与HTTP的Response的code混淆。
    """

    SUCCESS = 20000, "成功。"
    BAD_REQUEST = 40000, "请求错误。"
    NOT_FOUND = 40001, "未找到。"
    USER_DISABLED = 40002, "用户已停用。"
    FORBIDDEN = 40003, "用户无权访问。"
    PASSWORD_NOT_CORRECT = 40003, "密码不正确。"
    ALREADY_EXIST = 40005, "用户名已存在。"
    USER_STILL_ASSIGNED = 40006, "无法删除角色，因为仍有用户分配了该角色。"
    ROLE_ALREADY_EXIST = 40007, "角色名称已存在。"
    ACCOUNT_NOT_CORRECT = 40008, "用户名和密码不正确。"
    USER_NOT_FOUND = 40009, "用户未找到。"
    CREATE_USER_FAIL = 40010, "创建用户失败。"
    FUND_NOT_FOUND = 40011, "基金未找到。"
    ACTIVATE_CODE_NOT_FOUND = 40012, "激活码未找到。"
    ACTIVATE_CODE_USED = 40013, "激活码已使用。"
    DURATION_NOT_FOUND = 40014, "申请时长未找到。"
    INTERNAL_SERVER_ERROR = 50000, "服务器内部错误。"
    ILLEGAL_TOKEN = 50001, "非法令牌。"
    NO_REFRESH_TOKEN = 50002, "更新令牌不存在。"
    OTHER_CLIENTS_LOGGED_IN = 50003, "其他客户端已登录。"
    TOKEN_EXPIRED = 50004, "令牌已过期。"
    CONFIRM_URI_NOT_SET = 50005, "确认激活码URI未设置。"
    NOT_IMPLEMENT = 50006, "未实现。"

    def __new__(cls, value, message):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.message = message
        return obj

    def to_json(self):
        return {"code": self.value, "message": self.message}

    def to_http_status_code(self) -> int:
        """把内部值转换为HTTPResponse的有意义的status_code

        Returns:
            int: http status code
        """
        return math.floor(self._value_ / 100)


class PortalException(Exception):
    """portal模块的自定义业务异常。内部含有ErrCode枚举对象"""

    def __init__(self, err_code: PortalErrCode, original_exception: Exception = None):
        self._err_code = err_code
        self._original_exception = original_exception

    def get_err_code(self) -> PortalErrCode:
        return self._err_code

    def get_original_exception(self) -> Exception:
        return self._original_exception


class PortalBaseModel(BaseModel):
    # 配置禁用保护命名空间
    model_config = ConfigDict(protected_namespaces=())


def map_exception_to_err_code(ex: Exception) -> PortalErrCode:
    """将普通异常映射到 ErrCode 的逻辑。这里可以根据异常类型或消息做更细致的映射

    Args:
        ex (Exception):

    Returns:
        ErrCode:
    """
    return PortalErrCode.INTERNAL_SERVER_ERROR


def to_http_ex(err_code: PortalErrCode) -> HTTPException:
    """把ErrCode转换为HTTPException

    Args:
        err_code (ErrCode)

    Returns:
        HTTPException
    """
    return HTTPException(
        status_code=err_code.to_http_status_code(), detail=err_code.message
    )


def try_filter(full_traceback: str) -> str:
    """过滤异常堆栈信息，酌情去掉无意义的栈顶信息"""
    lines = full_traceback.splitlines()
    if len(lines) >= 8:
        return "\n".join(lines[4:])
    return full_traceback


def api_exception_handler(func):
    """REST API的装饰器。捕获Exception异常，转换为HTTPException

    Raises:
        HTTPException
    """

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except PortalException as myexc:
            full_traceback = traceback.format_exc()
            first_line = "Traceback (filtered most recent call)"
            logger.error(f"{myexc}\n{first_line}\n{try_filter(full_traceback)}")
            raise to_http_ex(myexc.get_err_code())
        except Exception as ex:
            full_traceback = traceback.format_exc()
            first_line = "Traceback (filtered most recent call)"
            logger.error(f"{ex}\n{first_line}\n{try_filter(full_traceback)}")
            raise to_http_ex(map_exception_to_err_code(ex))

    return wrapper


if __name__ == "__main__":
    print(PortalErrCode.SUCCESS.value, PortalErrCode.SUCCESS.message)
    print(PortalErrCode.ACCOUNT_NOT_CORRECT)



import os
import json
import importlib
from typing import Optional, Type
from .models import AllTaskTypes, AllCandidateTypes


class BaseSystem:
    """
    All systems should implement this interface.
    Arguments to the constructor must be serializable by Pydantic.
    System is assumed to be stateless.

    Subclasses must set self.candidate_class in their __init__ method.
    """

    def __init__(self, system_name: str):
        self.system_name = system_name
        # Subclasses must set this:
        self.candidate_class = None  # Type[AllCandidateTypes]

    def get_answer(
        self, task_id: str, task: AllTaskTypes, output_dir: str
    ) -> Optional[AllCandidateTypes]:
        """
        Return an answer for the question. Should use save_answer_to_disk to save the answer.

        Args:
            task_id (str): The ID of the task.
            task (AllTaskTypes): The typed task data.
            output_dir (str): The directory to save the output.

        Returns:
            Optional[AllCandidateTypes]: The typed candidate answer.
        """
        raise NotImplementedError("Implement your system's logic here.")

    def load_answer_from_disk(
        self, task_id: str, output_dir: str
    ) -> Optional[AllCandidateTypes]:
        """
        Helper to load an answer from disk if it exists.

        Args:
            task_id (str): The ID of the task.
            output_dir (str): The directory to load the answer from.

        Returns:
            Optional[AllCandidateTypes]: The loaded answer, or None if it doesn't exist.
        """
        if self.candidate_class is None:
            raise ValueError("Subclass must set self.candidate_class in __init__")

        answer_path = os.path.join(output_dir, f"{task_id}_answer.json")
        if not os.path.exists(answer_path):
            return None
        with open(answer_path, "r", encoding="utf-8") as f:
            data = json.load(f)
        return self.candidate_class.model_validate(data)

    def save_answer_to_disk(
        self, task_id: str, answer: AllCandidateTypes, output_dir: str
    ) -> None:
        """
        Save the answer to disk using Pydantic's json() method.

        Args:
            task_id (str): The ID of the task.
            answer (AllCandidateTypes): The typed candidate answer.
            output_dir (str): The directory to save the answer in.
        """
        os.makedirs(output_dir, exist_ok=True)
        answer_path = os.path.join(output_dir, f"{task_id}_answer.json")
        with open(answer_path, "w", encoding="utf-8") as f:
            f.write(answer.model_dump_json(indent=2))


def load_system_class(system_name: str) -> Type[BaseSystem]:
    """
    Dynamically load a system class based on the system name.

    Args:
        system_name (str): The name of the system.

    Returns:
        Type[BaseSystem]: The loaded system class.
    """
    module_name = "magentic_ui.eval.systems"
    class_name = f"{system_name}System"
    module = importlib.import_module(module_name)
    system_class = getattr(module, class_name)
    return system_class



import json
from typing import Optional, List, Dict, Sequence, Union, Any
from autogen_agentchat.messages import BaseAgentEvent
from pydantic import BaseModel
from dataclasses import dataclass
from pathlib import Path


@dataclass
class RunPaths:
    """
    A dataclass that contains the paths to the run directories.

    Attributes:
        internal_root_dir (Path): The base path for all potential runs.
        external_root_dir (Path): The base path for all potential runs.
        run_suffix (str): The suffix for the run directories.
        internal_run_dir (Path): The directory for the internal run.
        external_run_dir (Path): The directory for the external run.
    """

    internal_root_dir: Path
    external_root_dir: Path
    run_suffix: str
    internal_run_dir: Path
    external_run_dir: Path


class PlanStep(BaseModel):
    """
    A class representing a single step in a plan.

    Attributes:
        title (str): The title of the step.
        details (str): The description of the step.
        agent_name (str): The name of the agent responsible for this step.
    """

    title: str
    details: str
    agent_name: str


class Plan(BaseModel):
    """
    A class representing a plan consisting of multiple steps.

    Attributes:
        task (str, optional): The name of the task. Default: empty string
        steps (List[PlanStep]): A list of steps to complete the task.

    Example:
        plan = Plan(
            task="Open Website",
            steps=[PlanStep(title="Open Google", details="Go to the website google.com")]
        )
    """

    task: Optional[str]
    steps: Sequence[PlanStep]

    def __getitem__(self, index: int) -> PlanStep:
        return self.steps[index]

    def __len__(self) -> int:
        return len(self.steps)

    def __str__(self) -> str:
        """Return the string representation of the plan."""
        plan_str = ""
        if self.task is not None:
            plan_str += f"Task: {self.task}\n"
        for i, step in enumerate(self.steps):
            plan_str += f"{i}. {step.agent_name}: {step.title}\n   {step.details}\n"
        return plan_str

    @classmethod
    def from_list_of_dicts_or_str(
        cls, plan_dict: Union[List[Dict[str, str]], str, List[Any], Dict[str, Any]]
    ) -> Optional["Plan"]:
        """Load Plan from a list of dictionaries or a JSON string."""
        if isinstance(plan_dict, str):
            plan_dict = json.loads(plan_dict)
        if len(plan_dict) == 0:
            return None
        assert isinstance(plan_dict, (list, dict))

        task = None
        if isinstance(plan_dict, dict):
            task = plan_dict.get("task", None)
            plan_dict = plan_dict.get("steps", [])

        steps: List[PlanStep] = []
        for raw_step in plan_dict:
            if isinstance(raw_step, dict):
                step: dict[str, Any] = raw_step
                steps.append(
                    PlanStep(
                        title=step.get("title", "Untitled Step"),
                        details=step.get("details", "No details provided."),
                        agent_name=step.get("agent_name", "agent"),
                    )
                )
        return cls(task=task, steps=steps) if steps else None


class HumanInputFormat(BaseModel):
    """
    A class to represent and validate human input format.

    Attributes:
        content (str): The content of the input.
        accepted (bool, optional): Whether the input is accepted or not. Default: False
        plan (Plan, optional): A plan object.
    """

    content: str
    accepted: bool = False
    plan: Optional[Plan] = None

    @classmethod
    def from_str(cls, input_str: str) -> "HumanInputFormat":
        """Load HumanInputFormat from a string after validation."""
        try:
            data = json.loads(input_str)
            if not isinstance(data, dict):
                raise ValueError("Input string must be a JSON object")
        except (json.JSONDecodeError, ValueError):
            data = {"content": input_str}
        assert isinstance(data, dict)

        return cls(
            content=str(data.get("content", "")),  # type: ignore
            accepted=bool(data.get("accepted", False)),  # type: ignore
            plan=Plan.from_list_of_dicts_or_str(data.get("plan", [])),  # type: ignore
        )

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "HumanInputFormat":
        """Load HumanInputFormat from a dictionary after validation."""
        return cls(
            content=str(input_dict.get("content", "")),
            accepted=bool(input_dict.get("accepted", False)),
            plan=input_dict.get("plan", None),  # type: ignore
        )

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the input."""
        return self.model_dump()

    def to_str(self) -> str:
        """Return the string representation of the input."""
        return json.dumps(self.model_dump())


class CheckpointEvent(BaseAgentEvent):
    state: str
    content: str = "Checkpoint"
    metadata: Dict[str, str] = {"internal": "yes"}

    def to_text(self) -> str:
        return "Checkpoint"



import os
from typing import Dict, Any
from dotenv import load_dotenv
load_dotenv()
from loguru import logger
from hepai import HepAI
from hepai.types import APIKeyInfo

class PersonalKeyConfigFetcher:
    """获取个人密钥配置"""
    def __init__(self):
        self.service_mode = os.getenv("SERVICE_MODE")
        if self.service_mode == "PROD":
            
            admin_api_key = os.getenv("HEPAI_APP_ADMIN_API_KEY")  # 从环境变量中读取API-KEY
            base_url = "https://aiapi.ihep.ac.cn/apiv2"
            assert admin_api_key, "HEPAI_APP_ADMIN_API_KEY is not set, please set it in .env file"
            self.client = HepAI(api_key=admin_api_key, base_url=base_url)

    def get_personal_key(self, username: str) -> str:
        """获取个人密钥"""
        
        if self.service_mode == "PROD":
            api_key: APIKeyInfo = self.client.fetch_api_key(username=username)
            if not api_key or not api_key.api_key:
                raise ValueError(f"API key for user {username} not found.")
            return api_key.api_key
        else:
            api_key = os.getenv("HEPAI_API_KEY", "hepai_api_key_not_found")
            # assert api_key, "HEPAI_API_KEY not found, please add it in .env or environment variables in development mode"
            if api_key == "hepai_api_key_not_found":
                logger.warning("Using HEPAI_API_KEY in development mode, please set it in .env or environment variables in production mode")
            return api_key
        
    
    def get_default_config(self, username: str) -> Dict[str, Any]:
        """获取默认配置"""
        
        # "openai/gpt-4.1"
        personal_key = self.get_personal_key(username=username)
        default_model_configs = f"""model_config: &client
  provider: drsai.HepAIChatCompletionClient
  config:
    model: "deepseek-ai/deepseek-v3:671b"
    base_url: "https://aiapi.ihep.ac.cn/apiv2"
    api_key: "{personal_key}"
    max_retries: 10

coder_client: *client
orchestrator_client: *client
web_surfer_client: *client
file_surfer_client: *client
action_guard_client: *client
"""

        return {
            "cooperative_planning": True,
            "autonomous_execution": False,
            "allowed_websites": [],
            "max_actions_per_step": 5,
            "multiple_tools_per_call": False,
            "max_turns": 20,
            "approval_policy": "auto-conservative",
            "allow_for_replans": True,
            "do_bing_search": False,
            "websurfer_loop": False,
            "model_configs": default_model_configs,
            "retrieve_relevant_plans": "never"
        }




from typing import Dict, List, Any
import os, shutil
from fastapi import (
    APIRouter, 
    File, 
    UploadFile, 
    Depends, 
    Request,
    HTTPException,
    )
from fastapi.responses import FileResponse, HTMLResponse
import uuid

from ..initialization import AppInitializer
from ..deps import get_db
from ...datamodel.db import UserFiles

router = APIRouter()
initializer=None

def upload_to_filesystem(file_path: str, user_id: str) -> Dict[str, Any]:
    from openai import OpenAI
    from .....drsai_adapter.singleton import personal_key_config_fetcher as fetcher
                
    client = OpenAI(
        base_url="https://aiapi.ihep.ac.cn/apiv2",
        api_key= fetcher.get_personal_key(username=user_id)
    )

    file_obj = client.files.create(
        file=open(file_path, "rb"),
        purpose="user_data"
    )
    url = f"https://aiapi.ihep.ac.cn/apiv2/files/{file_obj.id}/preview"
    file_obj = file_obj.model_dump()
    file_obj["url"] = url
    return file_obj

@router.post("/")
async def upload_files(
    user_id: str, 
    session_id: str,
    files: List[UploadFile] = File(...), 
    db=Depends(get_db)
    ) -> Dict:
    '''
    接受上传的文件列表，解析上传到本地
    '''
    try:
        userfiles_path =  str(initializer.user_files / user_id / session_id)
        if not os.path.exists(userfiles_path):
            os.makedirs(userfiles_path, exist_ok=True)

        file_info = {} # 储存文件的名称、绝对路径、后缀名、byte大小

        # 保存文件到本地
        for file in files:
            
            # 首先判断文件大小是否超过了10MB
            if file.size > 10485760:
                raise HTTPException(status_code=413, detail="单个文件大小不能超过10MB，需要使用知识库进行上传：https://aiweb01.ihep.ac.cn:886/knowledge(Size limit exceeded 10MB)")
            
            file_path = os.path.join(userfiles_path, file.filename)
            file_id = str(uuid.uuid4())
            with open(file_path, "wb") as buffer:
                shutil.copyfileobj(file.file, buffer)
            
            # 顺便上传到文件系统
            # SERVICE_MODE = os.getenv("SERVICE_MODE", None)
            # if SERVICE_MODE == "PROD":
            #     file_obj = upload_to_filesystem(file_path, user_id)
            # else:
            #     file_obj = None

            file_info[file_id] = {
                "name": file.filename,
                "path": file_path,
                "suffix": os.path.splitext(file.filename)[1],
                "size": os.path.getsize(file_path),
                # "hepai_file_obj": file_obj
            }
        
        # 保存文件到数据库
        response = db.get(UserFiles, filters={"user_id": user_id, "session_id": session_id})
        if not response.status or not response.data:
            userfiles = UserFiles(
                user_id=user_id, 
                session_id=session_id,
                files=file_info,
                )
        else:
            # file_info_org: dict[str, Any] = response.data[0]["files"]
            # file_info_org.update(file_info)
            # file_info = file_info_org
            userfiles: UserFiles = response.data[0]
            if userfiles.files:
                userfiles.files.update(file_info)
            else:
                userfiles.files = file_info

        
        
        db.upsert(userfiles)
        return {"status": True, "data": file_info}
    except Exception as e:
        # Clean up session if run creation failed
        raise HTTPException(status_code=500, detail=str(e)) from e

# @router.get("/")
# async def get_user_files(user_id: str, db=Depends(get_db)) -> Dict:
#     """
#     检索用户上传的文件列表
#     """
#     response = db.get(UserFiles, filters={"user_id": user_id})
#     if not response.status or not response.data:
#         return {"status": False, "data": {}}
#     else:
#         # 处理多个session
#         user_files = {}
#         for files_i in response.data:
#             if "files" in files_i:
#                 user_files.update(files_i["files"])

#         return {"status": True, "data": response.data[0]["files"]}

@router.get("/{session_id}")
async def get_user_session_files(session_id: str, user_id: str, db=Depends(get_db)) -> Dict:
    """
    检索用户上传的文件列表
    """
    response = db.get(UserFiles, filters={"user_id": user_id, "session_id": session_id})
    if not response.status or not response.data:
        return {"status": False, "data": {}}
    else:
        userfiles: UserFiles = response.data[0]
        if not userfiles.files:
            return {"status": False, "data": userfiles.files}

        return {"status": False, "data": {}}


from fastapi import FastAPI, APIRouter, WebSocket, Request, Path, Query
from fastapi.responses import Response
import httpx
import websockets
import asyncio

app = FastAPI()
router = APIRouter()

@router.get("/{port}/{path:path}")
async def proxy_vnc(
    request: Request,
    port: int = Path(..., description="VNC server port"),
    path: str = Path(..., description="VNC server path"),
):
    VNC_HTTP_URL = f"http://localhost:{port}"
    url = f"{VNC_HTTP_URL}/{path}"
    try:
        async with httpx.AsyncClient() as client:
            vnc_response = await client.get(url, params=request.query_params)
        # 过滤掉部分 header
        excluded_headers = {"content-encoding", "transfer-encoding", "content-length", "connection"}
        headers = {k: v for k, v in vnc_response.headers.items() if k.lower() not in excluded_headers}
        return Response(content=vnc_response.content, 
                        status_code=vnc_response.status_code,
                        headers=headers)
    except httpx.ConnectError:
        return Response(content="Unable to connect to VNC backend.", status_code=502)
    except Exception as e:
        print(f"Proxy error: {e}")
        return Response(content="Internal server error.", status_code=500)

@router.websocket("/{port}/websockify")
async def websocket_proxy(websocket: WebSocket, port: int = Path(..., description="VNC server port")):
    VNC_WS_URL = f"ws://localhost:{port}"
    await websocket.accept()
    try:
        async with websockets.connect(VNC_WS_URL + "/websockify") as ws:
            consumer_task = asyncio.create_task(forward_messages(websocket, ws))
            producer_task = asyncio.create_task(forward_messages(ws, websocket))
            done, pending = await asyncio.wait(
                [consumer_task, producer_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
    except Exception as e:
        print(f"WebSocket connection error: {e}")
    finally:
        try:
            await websocket.close()
        except Exception:
            pass

async def forward_messages(src, dst):
    try:
        while True:
            # FastAPI WebSocket
            if hasattr(src, "receive"):
                data = await src.receive()
                if "bytes" in data and data["bytes"] is not None:
                    message = data["bytes"]
                    # websockets 库用 send 发送二进制
                    await dst.send(message)
                elif "text" in data and data["text"] is not None:
                    message = data["text"]
                    await dst.send(message)
            # websockets 库
            else:
                message = await src.recv()
                # 判断消息类型，转发给 FastAPI WebSocket
                if isinstance(message, bytes):
                    await dst.send_bytes(message)
                else:
                    await dst.send_text(message)
    except Exception as e:
        print(f"Error while forwarding messages: {e}")

app.include_router(router, prefix="/vncapi")


# /api/runs routes
from typing import Dict

from fastapi import APIRouter, Depends, HTTPException
from pydantic import BaseModel

from ...datamodel import Message, Run, RunStatus, Session
from ..deps import get_db

router = APIRouter()


class CreateRunRequest(BaseModel):
    session_id: int
    user_id: str


@router.post("/")
async def create_run(
    request: CreateRunRequest,
    db=Depends(get_db),
) -> Dict:
    """Return the existing run for a session or create a new one"""
    # First check if session exists and belongs to user
    session_response = db.get(
        Session,
        filters={"id": request.session_id, "user_id": request.user_id},
        return_json=False,
    )
    if not session_response.status or not session_response.data:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get the latest run for this session
    run_response = db.get(
        Run,
        filters={"session_id": request.session_id},
        return_json=False,
    )

    if not run_response.status or not run_response.data:
        # Create a new run if one doesn't exist
        try:
            run_response = db.upsert(
                Run(
                    session_id=request.session_id,
                    status=RunStatus.CREATED,
                    user_id=request.user_id,
                    task=None,
                    team_result=None,
                ),
                return_json=False,
            )
            if not run_response.status:
                raise HTTPException(status_code=400, detail=run_response.message)
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e)) from e

    # Return the run (either existing or newly created)
    run = None
    if isinstance(run_response.data, list):
        # get the run with the latest created_at
        run = max(run_response.data, key=lambda x: x.created_at)
    else:
        run = run_response.data
    return {"status": run_response.status, "data": {"run_id": str(run.id)}}


# We might want to add these endpoints:


@router.get("/{run_id}")
async def get_run(run_id: int, db=Depends(get_db)) -> Dict:
    """Get run details including task and result"""
    run = db.get(Run, filters={"id": run_id}, return_json=False)
    if not run.status or not run.data:
        raise HTTPException(status_code=404, detail="Run not found")

    return {"status": True, "data": run.data[0]}


@router.get("/{run_id}/messages")
async def get_run_messages(run_id: int, db=Depends(get_db)) -> Dict:
    """Get all messages for a run"""
    messages = db.get(
        Message, filters={"run_id": run_id}, order="created_at asc", return_json=False
    )

    return {"status": True, "data": messages.data}



from .approval_guard import (
    BaseApprovalGuard,
    MaybeRequiresApproval,
    DEFAULT_REQUIRES_APPROVAL,
)
from .tools.tool_metadata import get_tool_metadata, REQUIRE_APPROVAL_KEY
from autogen_core.tools import ToolSchema
from autogen_core.models import LLMMessage
from autogen_agentchat.messages import (
    MultiModalMessage,
    TextMessage,
)

from typing import (
    Any,
    Dict,
    List,
    Optional,
    cast,
    Protocol,
    TypeVar,
    Union,
    Callable,
    Generic,
)
from dataclasses import dataclass

from inspect import iscoroutinefunction

from abc import ABC, abstractmethod

TReturn = TypeVar("TReturn", covariant=True)
TStream = TypeVar("TStream", covariant=True)


class SyncActionCallable(Protocol[TReturn]):
    def __call__(self, *args: Any, **kwargs: Any) -> TReturn: ...


class AsyncActionCallable(Protocol[TReturn]):
    async def __call__(self, *args: Any, **kwargs: Any) -> TReturn: ...


ActionCallable = SyncActionCallable[TReturn] | AsyncActionCallable[TReturn]


class CallableInvoker(Generic[TReturn]):
    def __init__(self, callable: ActionCallable[TReturn]) -> None:
        self._callable = callable
        self._is_async = iscoroutinefunction(callable)

    async def __call__(self, *args: Any, **kwargs: Any) -> TReturn:
        if self._is_async:
            return await cast(AsyncActionCallable[TReturn], self._callable)(
                *args, **kwargs
            )
        else:
            return cast(SyncActionCallable[TReturn], self._callable)(*args, **kwargs)


class ApprovalDeniedError(Exception):
    """Exception raised when an action is denied by the approval guard."""

    ...


DescriptionGenerator = Callable[[Dict[str, Any]], Union[TextMessage, MultiModalMessage]]


@dataclass
class BaseGuardedAction(Generic[TReturn], ABC):
    name: str
    action: CallableInvoker[TReturn]
    prepare: Optional[CallableInvoker[None]] = None
    cleanup: Optional[CallableInvoker[None]] = None

    @abstractmethod
    def _get_baseline(self) -> MaybeRequiresApproval: ...

    async def invoke_with_approval(
        self,
        call_arguments: Dict[str, Any],
        action_description: Union[
            TextMessage,
            MultiModalMessage,
            ActionCallable[TextMessage]
            | ActionCallable[MultiModalMessage]
            | ActionCallable[TextMessage | MultiModalMessage],
        ],
        action_context: List[LLMMessage],
        action_guard: Optional[BaseApprovalGuard],
        action_description_for_user: Optional[
            Union[TextMessage, MultiModalMessage]
        ] = None,
    ) -> TReturn:
        """
        Invokes the action with approval if the action guard is provided.
        Args:
            call_arguments (Dict[str, Any]): The arguments to pass to the action.
            action_description (TextMessage | MultiModalMessage): The description of the action to be approved in it's raw form.
            action_context (List[LLMMessage]): The context of the action to be approved.
            action_guard (ApprovalGuard, optional): The action guard to use to approve the action.
            action_description_for_user (TextMessage | MultiModalMessage, optional): The description of the action for the user.
        """
        needs_approval: bool = False
        if action_guard is not None:
            baseline: MaybeRequiresApproval = self._get_baseline()
            llm_guess: MaybeRequiresApproval = baseline

            if REQUIRE_APPROVAL_KEY in call_arguments:
                if call_arguments[REQUIRE_APPROVAL_KEY]:
                    llm_guess = "always"
                else:
                    llm_guess = "never"

            # Check if the action needs approval
            needs_approval = await action_guard.requires_approval(
                baseline,
                llm_guess,
                action_context,
            )

        if self.prepare is not None:
            await self.prepare()

        try:
            if needs_approval:
                assert action_guard is not None

                if callable(action_description):
                    # If action_description is a callable, convert it to an Invoker and call it
                    invoker = CallableInvoker(action_description)
                    action_description = await invoker(**call_arguments)

                # Get approval for the action
                if action_description_for_user is None:
                    approved = await action_guard.get_approval(action_description)
                else:
                    approved = await action_guard.get_approval(
                        action_description_for_user
                    )

                if not approved:
                    raise ApprovalDeniedError(
                        "Action was denied by the approval guard."
                    )

            # Invoke the action
            result = await self.action(**call_arguments)

            if self.cleanup is not None:
                await self.cleanup()

            return result
        except Exception:
            if self.cleanup is not None:
                await self.cleanup()

            raise


class GuardedAction(BaseGuardedAction[TReturn], Generic[TReturn]):
    def __init__(
        self,
        name: str,
        action: ActionCallable[TReturn],
        prepare: Optional[ActionCallable[None]] = None,
        cleanup: Optional[ActionCallable[None]] = None,
    ) -> None:
        super().__init__(
            name,
            CallableInvoker(action),
            CallableInvoker(prepare) if prepare else None,
            CallableInvoker(cleanup) if cleanup else None,
        )

    @staticmethod
    def from_schema(
        tool_schema: ToolSchema,
        action: ActionCallable[TReturn],
        prepare: Optional[ActionCallable[None]] = None,
        cleanup: Optional[ActionCallable[None]] = None,
    ) -> "GuardedAction[TReturn]":
        return GuardedAction(
            name=tool_schema.get("name"),
            action=CallableInvoker(action),
            prepare=CallableInvoker(prepare) if prepare else None,
            cleanup=CallableInvoker(cleanup) if cleanup else None,
        )

    def _get_baseline(self) -> MaybeRequiresApproval:
        metadata = get_tool_metadata(self.name)

        return metadata.get("requires_approval", DEFAULT_REQUIRES_APPROVAL)


class TrivialGuardedAction(BaseGuardedAction[None]):
    def __init__(
        self, name: str, baseline_override: Optional[MaybeRequiresApproval] = None
    ) -> None:
        super().__init__(name, CallableInvoker(lambda: None), None, None)
        self._baseline_override: MaybeRequiresApproval | None = baseline_override

    def _get_baseline(self) -> MaybeRequiresApproval:
        return (
            self._baseline_override
            if self._baseline_override is not None
            else DEFAULT_REQUIRES_APPROVAL
        )



import json
import requests
from typing import List, Dict, Any


def download_file(url: str, filepath: str) -> None:
    response = requests.get(url)
    response.raise_for_status()
    with open(filepath, "wb") as file:
        file.write(response.content)


def load_jsonl(filepath: str) -> List[Dict[str, Any]]:
    with open(filepath, "r", encoding="utf-8") as file:
        return [json.loads(line) for line in file]


def load_json(filepath: str) -> Dict[str, Any]:
    with open(filepath, "r", encoding="utf-8") as file:
        return json.load(file)



import random
import string
from datetime import timedelta, datetime, UTC
from typing import Union, Optional

from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt, ExpiredSignatureError
from pydantic import BaseModel

# from haiddf.components.portal.common import PortalErrCode
# from haiddf.components.portal.constants import SECRET_KEY, ALGORITHM
# from haiddf.components.portal.models.data_models import UserInfoModel
# from haiddf.components.portal.models.view_models import AccessToken, AccessTokenData
from .common import PortalErrCode


class AccessToken(BaseModel):
    token_type: str = "Bearer"
    access_token: str


class AccessTokenData(BaseModel):
    username: Optional[str] = None


SECRET_KEY = "b6b14c922c4362377bf8d86e37847adca5ca1a3c54141bc58a10d8ea08f0971a"
ALGORITHM = "HS256"

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/apiv2/portal/user/login")

illegal_token_exception = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail=PortalErrCode.ILLEGAL_TOKEN.to_json(),
    headers={"WWW-Authenticate": "Bearer"},
)
expire_exception = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail=PortalErrCode.TOKEN_EXPIRED.to_json(),
    headers={"WWW-Authenticate": "Bearer"},
)


# async def authenticate_user(username: str, password: str) -> Union[UserInfoModel, bool]:
#     from haiddf.components.portal.singletons import userDao

#     user = await userDao.get_user_by_name(username)
#     if not user:
#         return False
#     # 这里是明文
#     if not password == user.password.get_secret_value():
#         return False
#     return user


def generate_random_string(length=5):
    """生成随机字符串"""
    letters = string.ascii_letters + string.digits
    return "".join(random.choice(letters) for i in range(length))


def create_jwt_token(data: dict, expires_delta: timedelta) -> AccessToken:
    to_encode = data.copy()
    begin = datetime.now(UTC)
    expire = begin + expires_delta

    to_encode.update({"exp": expire, "nonce": generate_random_string()})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return AccessToken(access_token=encoded_jwt)


def decode_jwt_token(token: str) -> AccessTokenData:
    try:
        payload = jwt.decode(
            token, SECRET_KEY, algorithms=[ALGORITHM], options={"verify_exp": True}
        )
        username: str = payload.get("sub")
        if username is None:
            raise illegal_token_exception
        return AccessTokenData(username=username)
    except ExpiredSignatureError:
        raise expire_exception
    except JWTError:
        raise illegal_token_exception


# async def get_current_user(token: str = Depends(oauth2_scheme)) -> UserInfoModel:
#     try:
#         payload = jwt.decode(
#             token, SECRET_KEY, algorithms=[ALGORITHM], options={"verify_exp": True}
#         )
#         username: str = payload.get("sub")
#         if username is None:
#             raise illegal_token_exception
#     except ExpiredSignatureError:
#         raise expire_exception
#     except JWTError:
#         raise illegal_token_exception

#     from haiddf.components.portal.singletons import userDao

#     user = await userDao.get_user_by_name(username)
#     if user is None:
#         raise illegal_token_exception

#     return user



import os
import hashlib
import base64
from typing import Optional, Dict, ClassVar
from loguru import logger
from pathlib import Path

from autogen_ext.experimental.task_centric_memory import (
    MemoryController,
    MemoryControllerConfig,
    MemoryBankConfig,
)
from autogen_ext.experimental.task_centric_memory.utils import PageLogger
from autogen_core.models import ChatCompletionClient


MEMORY_SUBDIR = "memory_bank"
LOG_SUBDIR = "pagelogs"


class MemoryControllerProvider:
    """Singleton provider for memory controller instances"""

    _instance: ClassVar[Optional["MemoryControllerProvider"]] = None
    _memory_controllers: Dict[str, MemoryController] = {}
    _internal_workspace_root: Optional[Path] = None
    _external_workspace_root: Optional[Path] = None
    _inside_docker: bool = False

    def __new__(
        cls,
        internal_workspace_root: Optional[Path] = None,
        external_workspace_root: Optional[Path] = None,
        inside_docker: bool = False,
    ):
        if cls._instance is None:
            cls._instance = super(MemoryControllerProvider, cls).__new__(cls)
            cls._instance._memory_controllers = {}
            cls._instance._internal_workspace_root = internal_workspace_root
            cls._instance._external_workspace_root = external_workspace_root
            cls._instance._inside_docker = inside_docker
        return cls._instance

    def __init__(
        self,
        internal_workspace_root: Optional[Path] = None,
        external_workspace_root: Optional[Path] = None,
        inside_docker: bool = False,
    ):
        """
        Initialize the memory controller provider with paths

        Args:
            internal_workspace_root (Path, optional): Path to workspace root inside docker
            external_workspace_root (Path, optional): Path to workspace root on host
            inside_docker (bool, optional): Whether code is running inside Docker. Default: False
        """

        if internal_workspace_root is not None:
            self._internal_workspace_root = internal_workspace_root

        if external_workspace_root is not None:
            self._external_workspace_root = external_workspace_root

        self._inside_docker = inside_docker

        root = (
            self._internal_workspace_root
            if self._inside_docker
            else self._external_workspace_root
        )
        if isinstance(root, Path):
            memory_dir = root / MEMORY_SUBDIR
            log_dir = root / LOG_SUBDIR

            memory_dir.mkdir(exist_ok=True, parents=True)
            log_dir.mkdir(exist_ok=True, parents=True)

            try:
                self.validate_path_safety(memory_dir, root)
                self.validate_path_safety(log_dir, root)
            except ValueError as e:
                logger.error(
                    f"Path safety validation failed during initialization: {e}"
                )
                raise ValueError(f"Security violation: {e}")

        else:
            logger.warning("Invalid root path: root must be a Path object")

    @staticmethod
    def get_safe_key(memory_controller_key: str) -> str:
        """Convert a user ID into a filesystem-safe key"""
        hash_obj = hashlib.sha256(memory_controller_key.encode("utf-8"))
        safe_id = (
            base64.urlsafe_b64encode(hash_obj.digest()[:16]).decode("utf-8").rstrip("=")
        )
        return safe_id

    @staticmethod
    def validate_path_safety(path: Path, base_dir: Path) -> bool:
        """
        Validate that a path is within the expected base directory.
        Helps prevent directory traversal attacks.

        Args:
            path (Path): The path to validate
            base_dir (Path): The base directory that should contain the path

        Returns:
            bool: True if path is safe

        Raises:
            ValueError: If the path is outside the base directory
        """
        resolved_path = os.path.realpath(path)
        resolved_base = os.path.realpath(base_dir)

        if not resolved_path.startswith(resolved_base):
            raise ValueError(f"Path validation failed: {path} is outside of {base_dir}")

        return True

    def get_path(self, subdir: str, safe_key: str) -> Optional[Path]:
        """Get path based on current context (Docker or host)"""
        root = (
            self._internal_workspace_root
            if self._inside_docker
            else self._external_workspace_root
        )

        if isinstance(root, Path):
            full_path = root / subdir / safe_key
            full_path.mkdir(exist_ok=True, parents=True)

            try:
                self.validate_path_safety(full_path, root / subdir)
                return full_path
            except ValueError:
                raise ValueError("Invalid memory controller key")

        logger.warning(
            "Memory controller provider root path is not a valid Path object"
        )
        return None

    def get_memory_controller(
        self,
        memory_controller_key: str,
        client: ChatCompletionClient,
        reset: bool = False,
    ) -> MemoryController:
        """Get or create a memory controller for the specified user"""
        safe_key = self.get_safe_key(memory_controller_key)

        if safe_key in self._memory_controllers and not reset:
            return self._memory_controllers[safe_key]

        memory_path = self.get_path(MEMORY_SUBDIR, safe_key)
        log_path = self.get_path(LOG_SUBDIR, safe_key)

        try:
            page_logger = PageLogger(config={"level": "INFO", "path": str(log_path)})

            memory_bank_config = MemoryBankConfig(
                path=str(memory_path),
                relevance_conversion_threshold=float(
                    os.environ.get("MEMORY_RELEVANCE_THRESHOLD", "1.7")
                ),
            )

            memory_controller_config = MemoryControllerConfig(
                generalize_task=False,
                revise_generalized_task=False,
                generate_topics=False,
                validate_memos=True,
                max_memos_to_retrieve=1,
                MemoryBank=memory_bank_config,
            )

            memory_controller = MemoryController(
                reset=reset,
                client=client,
                logger=page_logger,
                config=memory_controller_config,
            )

            self._memory_controllers[safe_key] = memory_controller
            return memory_controller

        except Exception as e:
            logger.error(f"Error creating memory controller: {e}")
            raise

    def close_memory_controller(self, memory_controller_key: str) -> None:
        """Close a memory controller and clean up resources"""
        safe_key = self.get_safe_key(memory_controller_key)
        if safe_key in self._memory_controllers:
            logger.info(f"Closing memory controller (safe key: {safe_key})")
            del self._memory_controllers[safe_key]

    def close_all_memory_controllers(self) -> None:
        """Close all memory controllers"""
        logger.info(
            f"Closing all memory controllers ({len(self._memory_controllers)} total)"
        )
        self._memory_controllers.clear()



import os
import logging
import shutil
import multiprocessing
import time
import json
import random
import datetime
from typing import Optional, Union, List, Tuple, Callable
from .benchmark import load_benchmark_class, Benchmark
from .basesystem import load_system_class, BaseSystem
from .models import AllCandidateTypes, AllEvalResultTypes


# ----------------------------------------------------------------------
# Setup Logging
# ----------------------------------------------------------------------
def _setup_file_logging(
    runs_dir: str,
    system_name: str,
    benchmark_name: str,
    split: Optional[str],
    run_id: Union[int, List[int]],
) -> None:
    """Setup file logging while maintaining console output.

    Args:
        runs_dir (str): Directory to store log file
        system_name (str): Name of the system
        benchmark_name (str): Name of the benchmark
        split (str, optional): Optional dataset split
        run_id (int | List[int]): Run ID or list of run IDs
    """
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    run_ids = [run_id] if isinstance(run_id, int) else run_id
    run_id_str = "-".join(str(rid) for rid in run_ids)

    log_dir = os.path.join(
        runs_dir,
        "runs",
        system_name,
        benchmark_name,
        split or "all_benchmark",
        str(run_id_str),
    )
    os.makedirs(log_dir, exist_ok=True)

    log_file = os.path.join(log_dir, f"core_eval_{timestamp}.log")

    # Create file handler
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s [%(levelname)s] %(name)s - %(message)s")
    )

    # Setup basic config for console output
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(name)s - %(message)s",
        handlers=[
            logging.StreamHandler(),  # Console handler
            file_handler,  # File handler
        ],
    )


logger = logging.getLogger(__name__)
logging.getLogger("autogen_core").setLevel(logging.CRITICAL)

# ----------------------------------------------------------------------
# Type Definitions & Constants
# ----------------------------------------------------------------------
# SystemType represents either a callable that creates a system or a system instance
SystemType = Union[Callable[..., BaseSystem], BaseSystem]
# BenchmarkType represents an optional callable that creates a benchmark
BenchmarkType = Optional[Callable[..., Benchmark]]


# ----------------------------------------------------------------------
# Helper Functions
# ----------------------------------------------------------------------
def _run_single_task(
    system_constructor: Union[Callable[[], BaseSystem], BaseSystem],
    task_id: str,
    output_dir: str,
    reload_system: bool,
    benchmark_constructor: Optional[Union[Callable[..., Benchmark], Benchmark]],
    benchmark_dir: str,
    reload_benchmark: bool,
    benchmark_name: str,
) -> Tuple[str, Optional[AllCandidateTypes], float]:
    """Run a single task in a separate process.

    Args:
        system_constructor (callable | BaseSystem): Callable or BaseSystem instance for creating/using the system
        task_id (str): Unique identifier for the task
        output_dir (str): Directory to store task outputs
        reload_system (bool): Whether to reload system for each task
        benchmark_constructor (callable, optional): Optional callable to create benchmark instance
        benchmark_dir (str): Directory containing benchmark data
        reload_benchmark (bool): Whether to reload benchmark for each task
        benchmark_name (str): Name of the benchmark

    Returns:
        Tuple containing:
            - str: Task identifier
            - Optional[AllCandidateTypes]: System's answer or None if failed
            - float: Time taken in seconds

    The function:
    1. Creates/reuses a system instance based on reload_system flag
    3. Checks for cached results to avoid re-running
    4. Executes the task and measures duration
    5. Saves results and timing information
    """
    logger.info(f"Running task {task_id} in {output_dir}")
    question_dir = os.path.join(output_dir, str(task_id))
    os.makedirs(question_dir, exist_ok=True)

    try:
        # Initialize or reload system
        if reload_system:
            assert callable(
                system_constructor
            ), "If reload_system is true, system_constructor must be callable"
            system = system_constructor()
        else:
            system = system_constructor

        assert isinstance(system, BaseSystem), "system must be a BaseSystem instance"

        # Initialize or reload benchmark if needed
        if isinstance(benchmark_constructor, Benchmark):
            benchmark = benchmark_constructor
        else:
            benchmark = download_and_load_benchmark(
                benchmark_name, benchmark_dir, benchmark_constructor
            )

        # Load task just before we need it
        task = benchmark.load_task_by_id(task_id)
        if task is None:
            raise ValueError(f"Task {task_id} not found")

        # If there's already an answer, skip
        if os.path.exists(question_dir):
            try:
                existing_answer = system.load_answer_from_disk(task_id, question_dir)
                if existing_answer:
                    times_path = os.path.join(question_dir, "times.json")
                    if os.path.exists(times_path):
                        with open(times_path, "r") as f:
                            times_data = json.load(f)
                            logger.info(f"Skipping {task_id} (already has answer).")
                            return (
                                task_id,
                                existing_answer,
                                times_data.get("duration", 0),
                            )
                    else:
                        raise FileNotFoundError(f"Times file not found for {task_id}")
            except Exception as e:
                logger.info(f"Error loading existing answer for {task_id}: {e}")
                # Clear question directory
                for file in os.listdir(question_dir):
                    file_path = os.path.join(question_dir, file)
                    if os.path.isfile(file_path):
                        os.remove(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)

        # Handle task files if they exist
        if hasattr(task, "file_name") and task.file_name:
            file_path = task.file_name
            if os.path.exists(file_path):
                shutil.copy(
                    file_path, os.path.join(question_dir, os.path.basename(file_path))
                )
            task.file_name = os.path.join(question_dir, os.path.basename(file_path))

        if hasattr(task, "file_dir") and task.file_dir:
            file_dir = task.file_dir
            if os.path.exists(file_dir):
                shutil.copytree(
                    file_dir, os.path.join(question_dir, os.path.basename(file_dir))
                )
            task.file_dir = os.path.join(question_dir, os.path.basename(file_dir))

        start_time = time.time()
        answer = system.get_answer(task_id, task, question_dir)
        end_time = time.time()

        times_path = os.path.join(question_dir, "times.json")
        with open(times_path, "w") as f:
            json.dump(
                {
                    "start_time": start_time,
                    "end_time": end_time,
                    "duration": end_time - start_time,
                },
                f,
            )

        logger.info(f"Completed task for task_id={task_id}")
        return task_id, answer, end_time - start_time
    except Exception as e:
        logger.info(f"Error running task for {task_id}: {e}")
        return task_id, None, 0


# ----------------------------------------------------------------------
# Core Functions
# ----------------------------------------------------------------------
def download_and_load_benchmark(
    benchmark_name: str,
    benchmark_dir: str,
    benchmark_constructor: Optional[Union[Callable[..., Benchmark], Benchmark]] = None,
) -> Benchmark:
    """Load or download benchmark data.

    Args:
        benchmark_name (str): Name of the benchmark
        benchmark_dir (str): Directory to store benchmark data
        benchmark_constructor (callable, optional): Optional callable to create benchmark instance

    Returns:
        Benchmark: Initialized benchmark instance

    The function:
    1. Creates benchmark instance using provided or default constructor
    2. Downloads benchmark data if not present locally
    3. Loads the dataset into memory
    """
    if benchmark_constructor is None:
        benchmark_class = load_benchmark_class(benchmark_name)
        benchmark_constructor = benchmark_class

    if callable(benchmark_constructor):
        data_dir = os.path.join(benchmark_dir, "data", benchmark_name)
        if not os.path.exists(data_dir):
            logger.info(f"Benchmark data not found in {data_dir}. Downloading...")
            os.makedirs(data_dir, exist_ok=True)
            benchmark = benchmark_constructor(name=benchmark_name, data_dir=data_dir)
            logger.info(f"Downloading benchmark {benchmark_name} into {data_dir}...")
            benchmark.download_dataset()
            logger.info("Download complete.")
        else:
            benchmark = benchmark_constructor(name=benchmark_name, data_dir=data_dir)
        benchmark.load_dataset()
        return benchmark
    else:
        return benchmark_constructor


def run_benchmark_func(
    benchmark_name: str,
    system_name: str,
    parallel: int,
    benchmark_dir: str,
    runs_dir: str,
    split: Optional[str] = None,
    run_id: int = 0,
    benchmark_constructor: Optional[Union[Callable[..., Benchmark], Benchmark]] = None,
    system_constructor: Optional[Union[Callable[..., BaseSystem], BaseSystem]] = None,
    subsample: Optional[float] = None,
    seed: Optional[int] = 42,
    reload_benchmark_per_task: bool = False,
    reload_system_per_task: bool = False,
) -> None:
    """Run benchmark evaluation.

    Args:
        benchmark_name (str): Name of the benchmark
        system_name (str): Name of the system to evaluate
        parallel (int): Number of parallel processes (1 for sequential)
        benchmark_dir (str): Directory containing benchmark data
        runs_dir (str): Directory to store run outputs
        split (str, optional): Optional dataset split to evaluate
        run_id (int, optional): Unique identifier for this run. Default: 0
        benchmark_constructor (Union[Callable[..., BaseSystem] | BaseSystem, optional): Optional callable to create benchmark instance
        system_constructor: Optional callable or instance to create/use system
        subsample (float, optional): Optional fraction (0-1] of tasks to evaluate
        seed (int, optional): Random seed for reproducibility. Default: 42
        reload_benchmark_per_task (bool, optional): Whether to reload benchmark for each task. Default: False
        reload_system_per_task (bool, optional): Whether to reload system for each task. Default: False

    The workflow:
    1. Sets up benchmark and system instances/constructors
    2. Prepares tasks with optional subsampling
    3. Executes tasks in parallel or sequentially
    4. Collects and logs results

    Key features:
    - Supports parallel processing
    - Handles system/benchmark reloading per task
    - Caches results to disk
    - Provides progress logging
    """
    _setup_file_logging(runs_dir, system_name, benchmark_name, split, run_id)
    if subsample is not None and not (0 < subsample <= 1):
        raise ValueError("subsample must be in the range (0, 1].")
    if seed is not None:
        random.seed(seed)

    # Validate reload_benchmark_per_task with parallel
    if reload_benchmark_per_task and parallel > 1:
        logger.info(
            "reload_benchmark_per_task=True is not supported in parallel mode. Setting to False."
        )
        reload_benchmark_per_task = False

    # Prepare benchmark constructor if needed
    if benchmark_constructor is None and benchmark_name:
        benchmark_class = load_benchmark_class(benchmark_name)
        benchmark_constructor = benchmark_class

    # Load initial benchmark instance
    benchmark = download_and_load_benchmark(
        benchmark_name, benchmark_dir, benchmark_constructor
    )

    output_dir = os.path.join(
        runs_dir,
        "runs",
        system_name,
        benchmark_name,
        split or "all_benchmark",
        str(run_id),
    )
    os.makedirs(output_dir, exist_ok=True)

    # System initialization can be done in three ways:
    # 1. Create new system for each task (reload_system_per_task=True)
    # 2. Reuse provided system instance
    # 3. Create single system instance and reuse
    if system_constructor is None:
        system_class = load_system_class(system_name)
        if reload_system_per_task:

            def create_system():
                return system_class(system_name)

            system_constructor = create_system
        else:
            system = system_class(system_name)
            system_constructor = system
    elif reload_system_per_task and callable(system_constructor):
        # Keep the constructor as is
        pass
    else:
        # Use the provided system directly
        system = system_constructor
        system_constructor = system

    # Get task IDs instead of full tasks
    task_ids = (
        benchmark.get_split_tasks(split) if split else list(benchmark.tasks.keys())
    )

    if subsample and 0 < subsample <= 1:
        task_ids = random.sample(task_ids, int(len(task_ids) * subsample))

    # Task preparation bundles all necessary data for parallel execution
    tasks_system_data = [
        (
            system_constructor,
            task_id,
            output_dir,
            reload_system_per_task,
            benchmark_constructor if reload_benchmark_per_task else benchmark,
            benchmark_dir,
            reload_benchmark_per_task,
            benchmark_name,
        )
        for task_id in task_ids
    ]

    logger.info(
        f"Starting run_benchmark with {'sequential' if parallel == 1 else str(parallel) + ' processes'}..."
    )

    # Separate path for non-parallel execution
    if parallel == 1:
        results: List[Tuple[str, Optional[AllCandidateTypes], float]] = []
        for task_data in tasks_system_data:
            results.append(_run_single_task(*task_data))
    else:
        with multiprocessing.Pool(processes=parallel) as pool:
            results = pool.starmap(_run_single_task, tasks_system_data)

    success_count = sum(1 for _, answer, _ in results if answer is not None)
    total_time = sum(t for _, a, t in results if a is not None)
    avg_time = total_time / success_count if success_count else 0
    logger.info(f"Average time per successful task: {avg_time:.4f} seconds")

    fail_count = len(results) - success_count

    logger.info(f"Run completed: {success_count} succeeded, {fail_count} failed.")


def _evaluate_single_task(
    task_id: str,
    system: BaseSystem,
    output_dir: str,
    benchmark: Benchmark,
    redo_eval: bool,
) -> Tuple[str, Optional[AllEvalResultTypes], float]:
    """Evaluate a single benchmark task.

    Args:
        task_id (str): Task identifier
        system (BaseSystem): System instance to evaluate
        output_dir (str): Directory containing system outputs
        benchmark (Benchmark): Benchmark instance for evaluation
        redo_eval (bool): Whether to redo evaluation

    Returns:
        Tuple containing:
            - str: Task identifier
            - Optional[AllEvalResultTypes]: Evaluation score or None if failed
            - float: Time taken in seconds

    The function:
    1. Loads the system's answer from disk
    2. Checks for existing evaluation scores
    3. Computes and caches the evaluation score
    4. Returns the score and timing information
    """
    logger.info(f"Evaluating task {task_id} in {output_dir}")
    question_dir = os.path.join(output_dir, str(task_id))
    times_path = os.path.join(question_dir, "times.json")
    score_path = os.path.join(question_dir, "score.json")

    if os.path.exists(times_path):
        with open(times_path, "r") as f:
            times_data = json.load(f)
            duration = times_data.get("duration", 0)
    else:
        duration = -1

    if not redo_eval and os.path.exists(score_path):
        try:
            with open(score_path, "r") as f:
                file_score = json.load(f)
                assert benchmark.eval_result_class is not None
                saved_score = benchmark.eval_result_class.model_validate(file_score)
                logger.info(f"Loaded existing score for {task_id}")
            return (task_id, saved_score, duration)
        except Exception as e:
            logger.info(f"Error loading existing score for {task_id}: {e}")

    # Load task just before evaluation
    task = benchmark.load_task_by_id(task_id)
    if task is None:
        logger.info(f"Task {task_id} not found")
        return (task_id, None, duration)

    candidate = system.load_answer_from_disk(task_id, question_dir)
    if candidate is None:
        logger.info(f"No candidate found for {task_id}")
        return (task_id, None, duration)
    logger.info(f"Evaluating candidate for {task_id}")
    score = benchmark.evaluator(task, candidate)
    logger.info(f"Finished evaluating candidate for {task_id} with score {score}")
    with open(score_path, "w") as f:
        json.dump(score.model_dump(), f)
    return (task_id, score, duration)


def evaluate_benchmark_func(
    benchmark_name: str,
    system_name: str,
    benchmark_dir: str,
    runs_dir: str,
    split: Optional[str] = None,
    run_id: Union[int, List[int]] = 0,
    benchmark_constructor: Optional[Union[Callable[..., Benchmark], Benchmark]] = None,
    system_constructor: Optional[Union[Callable[..., BaseSystem], BaseSystem]] = None,
    parallel: int = 1,
    redo_eval: bool = False,
) -> None:
    """Evaluates benchmark results across single or multiple runs.
    Args:
        benchmark_name (str): Name of the benchmark
        system_name (str): Name of the system evaluated
        benchmark_dir (str): Directory containing benchmark data
        runs_dir (str): Directory containing run outputs
        split (str, optional): Optional dataset split that was evaluated
        run_id (int, optional): Run ID or list of run IDs to evaluate. Default: 0
        benchmark_constructor (callable, optional): Optional callable to create benchmark instance
        system_constructor (callable | BaseSystem, optional): Optional callable or instance to create/use system
        parallel (int, optional): Number of parallel processes (1 for sequential). Default: 1
        redo_eval (bool, optional): Whether to redo evaluation even if results exist. Default: False

    The workflow:
    1. Processes each run ID separately
    2. Loads or creates necessary system/benchmark instances
    3. Evaluates all task results
    4. Computes and saves aggregate metrics

    Features:
    - Supports evaluating multiple runs
    - Parallel evaluation processing
    - Caches evaluation scores
    - Computes aggregate metrics across runs
    """
    _setup_file_logging(runs_dir, system_name, benchmark_name, split, run_id)
    if isinstance(run_id, int):
        run_ids = [run_id]
    else:
        run_ids = run_id

    all_scores: List[List[AllEvalResultTypes]] = []
    all_durations: List[List[float]] = []
    all_quids: List[List[str]] = []
    for rid in run_ids:
        # Prepare benchmark constructor if needed
        if benchmark_constructor is None and benchmark_name:
            benchmark_class = load_benchmark_class(benchmark_name)
            benchmark_constructor = benchmark_class

        benchmark = download_and_load_benchmark(
            benchmark_name, benchmark_dir, benchmark_constructor
        )

        output_dir = os.path.join(
            runs_dir,
            "runs",
            system_name,
            benchmark_name,
            split or "all_benchmark",
            str(rid),
        )
        if not os.path.exists(output_dir):
            raise FileNotFoundError(
                f"No system output found at {output_dir}. Run the benchmark first."
            )

        # Initialize system
        if system_constructor is None:
            system_class = load_system_class(system_name)
            system = system_class(system_name)
        elif callable(system_constructor):
            system = system_constructor()
        else:
            system = system_constructor

        tasks = benchmark.get_split_tasks(split) if split else benchmark.tasks
        tasks_sys_benchmark_data = [
            (task_id, system, output_dir, benchmark, redo_eval) for task_id in tasks
        ]

        # Separate path for non-parallel execution
        if parallel == 1:
            single_results: List[Tuple[str, Optional[AllEvalResultTypes], float]] = []
            for task_sys_benchmark_data in tasks_sys_benchmark_data:
                single_results.append(_evaluate_single_task(*task_sys_benchmark_data))
        else:
            with multiprocessing.Pool(processes=parallel) as pool:
                single_results = pool.starmap(
                    _evaluate_single_task, tasks_sys_benchmark_data
                )

        # Process single_results in place of the for ex in exs loop
        scores: List[AllEvalResultTypes] = []
        durations: List[float] = []
        quids: List[str] = []
        failed_count = 0
        for qid, score, duration in single_results:
            if score is not None:
                scores.append(score)
                quids.append(qid)
                if duration != -1:
                    durations.append(duration)
            else:
                failed_count += 1

        logger.info(
            f"Evaluation results for run {rid}: {len(scores)} successful, {failed_count} failed/missing"
        )

        if durations:
            avg_time = sum(durations) / len(durations)
            logger.info(f"Average time across evaluated tasks: {avg_time:.4f} s")
        else:
            avg_time = -1

        metrics = benchmark.compute_aggregate_metrics(scores)
        logger.info(f"Evaluation metrics: {metrics}")

        # Add average time and scores to metrics
        metrics["average_time"] = avg_time
        metrics["scores"] = [
            (id, score.model_dump_json()) for id, score in zip(quids, scores)
        ]

        # Save metrics to a file
        metrics_path = os.path.join(output_dir, "metrics.json")
        with open(metrics_path, "w") as f:
            json.dump(metrics, f)
        logger.info(f"Metrics saved to {metrics_path}")

        all_scores.append(scores)
        all_durations.append(durations)
        all_quids.append(quids)
    if len(run_ids) > 1:
        benchmark = download_and_load_benchmark(
            benchmark_name, benchmark_dir, benchmark_constructor
        )
        aggregate_metrics = benchmark.compute_aggregate_metrics_multiple_runs(
            all_scores, all_durations
        )

        # Save aggregate metrics to a file
        aggregate_metrics_path = os.path.join(
            runs_dir,
            "runs",
            system_name,
            benchmark_name,
            split or "all_benchmark",
            "aggregate_metrics.json",
        )
        with open(aggregate_metrics_path, "w") as f:
            json.dump(aggregate_metrics, f)
        logger.info(f"Aggregate metrics saved to {aggregate_metrics_path}")


def run_evaluate_benchmark_func(
    benchmark_name: str,
    system_name: str,
    parallel: int,
    benchmark_dir: str,
    runs_dir: str,
    split: Optional[str] = None,
    run_id: Union[int, List[int]] = 0,
    benchmark_constructor: Optional[Union[Callable[..., Benchmark], Benchmark]] = None,
    system_constructor: Optional[Union[Callable[..., BaseSystem], BaseSystem]] = None,
    subsample: Optional[float] = None,
    seed: Optional[int] = None,
    redo_eval: bool = False,
    reload_benchmark_per_task: bool = False,
    reload_system_per_task: bool = False,
) -> None:
    """Run benchmark evaluation and compute metrics.

    Args:
        benchmark_name (str): Name of the benchmark
        system_name (str): Name of the system to evaluate
        parallel (int): Number of parallel processes
        benchmark_dir (str): Directory containing benchmark data
        runs_dir (str): Directory to store outputs
        split (str, optional): Optional dataset split to evaluate
        run_id (int, List[int]): Run ID or list of run IDs. Default: 0
        benchmark_constructor (callable, optional): Optional callable to create benchmark instance
        system_constructor (callable | BaseSystem, optional): Optional callable or instance to create/use system
        subsample (float, optional): Optional fraction (0-1] of tasks to evaluate
        seed (int, optional): Optional random seed for reproducibility
        redo_eval (bool, optional): Whether to redo evaluation even if results exist. Default: False
        reload_benchmark_per_task (bool, optional): Whether to reload benchmark for each task. Default: False
        reload_system_per_task (bool, optional): Whether to reload system for each task. Default: False
    """
    if isinstance(run_id, int):
        run_ids = [run_id]
    else:
        run_ids = run_id

    for rid in run_ids:
        run_benchmark_func(
            benchmark_name=benchmark_name,
            system_name=system_name,
            parallel=parallel,
            benchmark_dir=benchmark_dir,
            runs_dir=runs_dir,
            split=split,
            run_id=rid,
            benchmark_constructor=benchmark_constructor,
            system_constructor=system_constructor,
            subsample=subsample,
            seed=seed,
            reload_benchmark_per_task=reload_benchmark_per_task,
            reload_system_per_task=reload_system_per_task,
        )
    evaluate_benchmark_func(
        benchmark_name=benchmark_name,
        system_name=system_name,
        benchmark_dir=benchmark_dir,
        runs_dir=runs_dir,
        split=split,
        run_id=run_ids,
        benchmark_constructor=benchmark_constructor,
        system_constructor=system_constructor,
        parallel=parallel,
        redo_eval=redo_eval,
    )
